import pytest
from baron import parse
from baron.render import render, node_types, nodes_rendering_order, RenderWalker


@pytest.fixture(params=nodes_rendering_order.keys())
def dictionnary_node(request):
    return nodes_rendering_order[request.param]


def test_render_crap():
    with pytest.raises(NotImplementedError):
        render("crap")


def test_dictionnary_key_validity(dictionnary_node):
    for key_type, _, _ in dictionnary_node:
        assert key_type in node_types


def test_dictionnary_dependent_validity(dictionnary_node):
    keys = set([t[1] for t in dictionnary_node])
    for key_type, _, dependent in dictionnary_node:
        assert isinstance(dependent, bool) \
            or (isinstance(dependent, str) and dependent in keys) \
            or (isinstance(dependent, list) and all([d in keys for d in dependent]))

        if key_type == 'bool':
            assert dependent is False


def test_render_dictionnary_bad_type():
    nodes_rendering_order['bad_type'] = [('wtf', 'hello', True)]
    with pytest.raises(NotImplementedError) as e:
        list(render({'type': 'bad_type'}))
    assert str(e.value) == "Unknown key type \"wtf\" in \"bad_type\" node"


def test_render_dictionnary_bad_bool_dependency():
    nodes_rendering_order['bad_bool_dependency'] = [('bool', True, True)]
    with pytest.raises(NotImplementedError) as e:
        list(render({'type': 'bad_bool_dependency'}))
    assert str(e.value) == "Bool keys are only used for dependency, they cannot be rendered. Please set the \"('bool', True, True)\"'s dependent key in \"bad_bool_dependency\" node to False"


def test_render_dictionnary_bad_bool_dependency2():
    nodes_rendering_order['bad_bool_dependency2'] = [('bool', False, 'other_key')]
    with pytest.raises(NotImplementedError) as e:
        list(render({'type': 'bad_bool_dependency2'}))
    assert str(e.value) == "Bool keys are only used for dependency, they cannot be rendered. Please set the \"('bool', False, 'other_key')\"'s dependent key in \"bad_bool_dependency2\" node to False"


class RenderWalkerTester(RenderWalker):
    def __init__(self, steps):
        super(RenderWalkerTester, self).__init__(strict=True)
        self.steps = steps

    def before(self, *args):
        super(RenderWalkerTester, self).before(*args)
        return self.process_test('>', *args)

    def after(self, *args):
        stop = self.process_test('<', *args)
        super(RenderWalkerTester, self).after(*args)
        return stop

    def process_test(self, direction, node_type, node, render_key):
        _direction, _node_type, _node, _render_key, _stop = self.steps.pop(0)
        target = (_direction, _node_type, _node, _render_key)
        if node_type in ['constant', 'string']:
            assert target == (direction, node_type, node, render_key)
        elif "type" in node:
            assert target == (direction, node_type, node["type"], render_key)
        else:
            assert target == (direction, node_type, node.__class__.__name__, render_key)
        return _stop


# flake8: noqa
def test_walk_stop():
    node = parse("a = 1")
    walker = RenderWalkerTester([
    ('>', 'node', 'assignment', 0, False),
        ('>', 'key', 'name', 'target', False),
            ('>', 'string', 'a', 'value', True),
        ('<', 'key', 'name', 'target', False),
    ('<', 'node', 'assignment', 0, False),
    ])

    walker.walk(node)


def test_walk_assignment():
    node = parse("a = 1")
    walker = RenderWalkerTester([
    ('>', 'node', 'assignment', 0, False),
        ('>', 'key', 'name', 'target', False),
            ('>', 'string', 'a', 'value', False),
            ('<', 'string', 'a', 'value', False),
        ('<', 'key', 'name', 'target', False),
        ('>', 'formatting', 'list', 'first_formatting', False),
            ('>', 'node', 'space', 0, False),
                ('>', 'string', ' ', 'value', False),
                ('<', 'string', ' ', 'value', False),
            ('<', 'node', 'space', 0, False),
        ('<', 'formatting', 'list', 'first_formatting', False),
        ('>', 'constant', '=', '=', False),
        ('<', 'constant', '=', '=', False),
        ('>', 'formatting', 'list', 'second_formatting', False),
            ('>', 'node', 'space', 0, False),
                ('>', 'string', ' ', 'value', False),
                ('<', 'string', ' ', 'value', False),
            ('<', 'node', 'space', 0, False),
        ('<', 'formatting', 'list', 'second_formatting', False),
        ('>', 'key', 'int', 'value', False),
            ('>', 'string', '1', 'value', False),
            ('<', 'string', '1', 'value', False),
        ('<', 'key', 'int', 'value', False),
    ('<', 'node', 'assignment', 0, False),
    ])

    walker.walk(node)


def test_walk_funcdef_with_leading_space():
    node = parse("""\

@deco
def fun(arg1):
    pass
""")
    walker = RenderWalkerTester([
    ('>', 'node', 'endl', 0, False),
        ('>', 'formatting', 'list', 'formatting', False),
        ('<', 'formatting', 'list', 'formatting', False),
        ('>', 'string', '\n', 'value', False),
        ('<', 'string', '\n', 'value', False),
        ('>', 'string', '', 'indent', False),
        ('<', 'string', '', 'indent', False),
    ('<', 'node', 'endl', 0, False),
    ('>', 'node', 'def', 1, False),
        ('>', 'list', 'list', 'decorators', False),
            ('>', 'node', 'decorator', 0, False),
                ('>', 'constant', '@', '@', False),
                ('<', 'constant', '@', '@', False),
                ('>', 'key', 'dotted_name', 'value', False),
                    ('>', 'list', 'list', 'value', False),
                        ('>', 'node', 'name', 0, False),
                            ('>', 'string', 'deco', 'value', False),
                            ('<', 'string', 'deco', 'value', False),
                        ('<', 'node', 'name', 0, False),
                    ('<', 'list', 'list', 'value', False),
                ('<', 'key', 'dotted_name', 'value', False),
            ('<', 'node', 'decorator', 0, False),
            ('>', 'node', 'endl', 1, False),
                ('>', 'formatting', 'list', 'formatting', False),
                ('<', 'formatting', 'list', 'formatting', False),
                ('>', 'string', '\n', 'value', False),
                ('<', 'string', '\n', 'value', False),
                ('>', 'string', '', 'indent', False),
                ('<', 'string', '', 'indent', False),
            ('<', 'node', 'endl', 1, False),
        ('<', 'list', 'list', 'decorators', False),
        ('>', 'constant', 'def', 'def', False),
        ('<', 'constant', 'def', 'def', False),
        ('>', 'formatting', 'list', 'first_formatting', False),
            ('>', 'node', 'space', 0, False),
                ('>', 'string', ' ', 'value', False),
                ('<', 'string', ' ', 'value', False),
            ('<', 'node', 'space', 0, False),
        ('<', 'formatting', 'list', 'first_formatting', False),
        ('>', 'string', 'fun', 'name', False),
        ('<', 'string', 'fun', 'name', False),
        ('>', 'formatting', 'list', 'second_formatting', False),
        ('<', 'formatting', 'list', 'second_formatting', False),
        ('>', 'constant', '(', '(', False),
        ('<', 'constant', '(', '(', False),
        ('>', 'formatting', 'list', 'third_formatting', False),
        ('<', 'formatting', 'list', 'third_formatting', False),
        ('>', 'list', 'list', 'arguments', False),
            ('>', 'node', 'def_argument', 0, False),
                ('>', 'key', 'name', 'target', False),
                    ('>', 'string', 'arg1', 'value', False),
                    ('<', 'string', 'arg1', 'value', False),
                ('<', 'key', 'name', 'target', False),
            ('<', 'node', 'def_argument', 0, False),
        ('<', 'list', 'list', 'arguments', False),
        ('>', 'formatting', 'list', 'fourth_formatting', False),
        ('<', 'formatting', 'list', 'fourth_formatting', False),
        ('>', 'constant', ')', ')', False),
        ('<', 'constant', ')', ')', False),
        ('>', 'formatting', 'list', 'fifth_formatting', False),
        ('<', 'formatting', 'list', 'fifth_formatting', False),
        ('>', 'constant', ':', ':', False),
        ('<', 'constant', ':', ':', False),
        ('>', 'formatting', 'list', 'sixth_formatting', False),
        ('<', 'formatting', 'list', 'sixth_formatting', False),
        ('>', 'list', 'list', 'value', False),
            ('>', 'node', 'endl', 0, False),
                ('>', 'formatting', 'list', 'formatting', False),
                ('<', 'formatting', 'list', 'formatting', False),
                ('>', 'string', '\n', 'value', False),
                ('<', 'string', '\n', 'value', False),
                ('>', 'string', '    ', 'indent', False),
                ('<', 'string', '    ', 'indent', False),
            ('<', 'node', 'endl', 0, False),
            ('>', 'node', 'pass', 1, False),
                ('>', 'string', 'pass', 'type', False),
                ('<', 'string', 'pass', 'type', False),
            ('<', 'node', 'pass', 1, False),
            ('>', 'node', 'endl', 2, False),
                ('>', 'formatting', 'list', 'formatting', False),
                ('<', 'formatting', 'list', 'formatting', False),
                ('>', 'string', '\n', 'value', False),
                ('<', 'string', '\n', 'value', False),
                ('>', 'string', '', 'indent', False),
                ('<', 'string', '', 'indent', False),
            ('<', 'node', 'endl', 2, False),
        ('<', 'list', 'list', 'value', False),
    ('<', 'node', 'def', 1, False),
    ])

    walker.walk(node)



from . import grouper  # noqa
from . import spliter  # noqa
from .baron import parse, tokenize  # noqa
from .dumper import dumps  # noqa
from .inner_formatting_grouper import GroupingError, UnExpectedFormattingToken  # noqa
from .parser import ParsingError  # noqa
from .render import nodes_rendering_order  # noqa
from .spliter import UntreatedError  # noqa
from .utils import BaronError  # noqa



from .utils import FlexibleIterator

"""
Objective: add an INDENT token and a DEDENT token arround every block.

Strategy: put after every ":" that is not in a slice/dictionary declaration/lambda.

Slice and dictionary are easy: increase a number when a "[" or "{" is found,
decrease it for a "]" or "}". If the number is != 0, we are in a dictionary or
slice -> do not put a INDENT when a ":" is found.

Lambda are a bit different: increase another number when a "lambda" is found,
if the number is != 0 and a ":" is found, decrease this number, otherwise put a
INDENT.

For the DEDENT, I'm probably going to need to keep a list of indentation and
decheck the last one every time I encounter a meaningfull line. Still need to
test this idea.
"""


def mark_indentation(sequence):
    return list(mark_indentation_generator(sequence))


def transform_tabs_to_spaces(string):
    return string.replace("\t", " " * 8)


def get_space(node):
    """ Return space formatting information of node.

    If the node does not have a third formatting item - like in
    a ('ENDL', '\n') node - then we return None as a flag value. This is
    maybe not the best behavior but it seems to work for now.
    """
    if len(node) < 4 or len(node[3]) == 0:
        return None
    return transform_tabs_to_spaces(node[3][0][1])


def mark_indentation_generator(sequence):
    iterator = FlexibleIterator(sequence)
    current = None, None
    indentations = []
    while True:
        if iterator.end():
            return

        current = next(iterator)

        if current is None:
            return

        # end of the file, I need to pop all indentations left and put the
        # corresponding dedent token for them
        if current[0] == "ENDMARKER" and indentations:
            while len(indentations) > 0:
                yield ('DEDENT', '')
                indentations.pop()

        # if were are at ":\n" like in "if stuff:\n"
        if current[0] == "COLON" and iterator.show_next(1)[0] == "ENDL":
            # if we aren't in "if stuff:\n\n"
            if iterator.show_next(2)[0] not in ("ENDL",):
                indentations.append(get_space(iterator.show_next()))
                yield current
                yield next(iterator)
                yield ('INDENT', '')
                continue
            else:  # else, skip all "\n"
                yield current
                for i in iterator:
                    if i[0] == 'ENDL' and iterator.show_next()[0] not in ('ENDL',):
                        indentations.append(get_space(i))
                        yield ('INDENT', '')
                        yield i
                        break
                    yield i
                continue

        # if we were in an indented situation and that the next line has a lower indentation
        if indentations and current[0] == "ENDL":
            the_indentation_level_changed = get_space(current) is None or get_space(current) != indentations[-1]
            if the_indentation_level_changed and iterator.show_next()[0] not in ("ENDL", "COMMENT"):
                new_indent = get_space(current) if len(current) == 4 else ""
                yield current

                # pop until reaching the matching indentation level
                while indentations and string_is_bigger(indentations[-1], new_indent):
                    indentations.pop()
                    yield ('DEDENT', '')
                yield next(iterator)
                continue

        yield current


def string_is_bigger(s1, s2):
    """ Return s1 > s2 by taking into account None values.

    None is always smaller than any string.

    None > "string" works in python2 but not in python3. This function
    makes it work in python3 too.
    """
    if s1 is None:
        return False
    elif s2 is None:
        return True
    else:
        return s1 > s2



import json
import sys
from os import linesep
from . import parse


def show(source_code):
    sys.stdout.write(json.dumps(parse(source_code), indent=4) + linesep)


def show_file(target_file):
    with open(target_file, "r") as source_code:
        sys.stdout.write(json.dumps(parse(source_code.read()), indent=4) + linesep)


def show_node(node):
    sys.stdout.write(json.dumps(node, indent=4) + linesep)



import re
from .utils import BaronError


class UnknowItem(BaronError):
    pass


KEYWORDS = ("and", "as", "assert", "break", "class", "continue", "def", "del",
            "elif", "else", "except", "exec", "finally", "for", "from",
            "global", "nonlocal", "if", "import", "in", "is", "lambda", "not",
            "or", "pass", "print", "raise", "return", "try", "while", "with",
            "yield")

TOKENS = (
    (r'[a-zA-Z_]\w*', 'NAME'),
    (r'0', 'INT'),
    (r'[-+]?\d+[eE][-+]?\d+[jJ]', 'FLOAT_EXPONANT_COMPLEX'),
    (r'[-+]?\d+.\d?[eE][-+]?\d+[jJ]', 'FLOAT_EXPONANT_COMPLEX'),
    (r'[-+]?\d?.\d+[eE][-+]?\d+[jJ]', 'FLOAT_EXPONANT_COMPLEX'),
    (r'\d+[eE][-+]?\d*', 'FLOAT_EXPONANT'),
    (r'\d+\.\d*[eE][-+]?\d*', 'FLOAT_EXPONANT'),
    (r'\.\d+[eE][-+]?\d*', 'FLOAT_EXPONANT'),
    (r'\d*\.\d+[jJ]', 'COMPLEX'),
    (r'\d+\.[jJ]', 'COMPLEX'),
    (r'\d+[jJ]', 'COMPLEX'),
    (r'\d+\.', 'FLOAT'),
    (r'\d*[_\d]*\.[_\d]+[lL]?', 'FLOAT'),
    (r'\d+[_\d]+\.[_\d]*[lL]?', 'FLOAT'),
    (r'\.', 'DOT'),
    (r'[1-9]+[_\d]*[lL]', 'LONG'),
    (r'[1-9]+[_\d]*', 'INT'),
    (r'0[xX][\d_a-fA-F]+[lL]?', 'HEXA'),
    (r'(0[oO][0-7]+)|(0[0-7_]*)[lL]?', 'OCTA'),
    (r'0[bB][01_]+[lL]?', 'BINARY'),
    (r'\(', 'LEFT_PARENTHESIS'),
    (r'\)', 'RIGHT_PARENTHESIS'),
    (r':', 'COLON'),
    (r',', 'COMMA'),
    (r';', 'SEMICOLON'),
    (r'@', 'AT'),
    (r'\+', 'PLUS'),
    (r'-', 'MINUS'),
    (r'\*', 'STAR'),
    (r'/', 'SLASH'),
    (r'\|', 'VBAR'),
    (r'&', 'AMPER'),
    (r'@', 'AT'),
    (r'<', 'LESS'),
    (r'>', 'GREATER'),
    (r'=', 'EQUAL'),
    (r'%', 'PERCENT'),
    (r'\[', 'LEFT_SQUARE_BRACKET'),
    (r'\]', 'RIGHT_SQUARE_BRACKET'),
    (r'\{', 'LEFT_BRACKET'),
    (r'\}', 'RIGHT_BRACKET'),
    (r'`', 'BACKQUOTE'),
    (r'==', 'EQUAL_EQUAL'),
    (r'<>', 'NOT_EQUAL'),
    (r'!=', 'NOT_EQUAL'),
    (r'<=', 'LESS_EQUAL'),
    (r'>=', 'GREATER_EQUAL'),
    (r'~', 'TILDE'),
    (r'\^', 'CIRCUMFLEX'),
    (r'<<', 'LEFT_SHIFT'),
    (r'>>', 'RIGHT_SHIFT'),
    (r'\*\*', 'DOUBLE_STAR'),
    (r'\+=', 'PLUS_EQUAL'),
    (r'-=', 'MINUS_EQUAL'),
    (r'@=', 'AT_EQUAL'),
    (r'\*=', 'STAR_EQUAL'),
    (r'/=', 'SLASH_EQUAL'),
    (r'%=', 'PERCENT_EQUAL'),
    (r'&=', 'AMPER_EQUAL'),
    (r'\|=', 'VBAR_EQUAL'),
    (r'\^=', 'CIRCUMFLEX_EQUAL'),
    (r'<<=', 'LEFT_SHIFT_EQUAL'),
    (r'>>=', 'RIGHT_SHIFT_EQUAL'),
    (r'\.\.\.', 'ELLIPSIS'),
    (r'->', 'RIGHT_ARROW'),
    (r'\*\*=', 'DOUBLE_STAR_EQUAL'),
    (r'//', 'DOUBLE_SLASH'),
    (r'//=', 'DOUBLE_SLASH_EQUAL'),
    (r'\n', 'ENDL'),
    (r'\r\n', 'ENDL'),
    (r'#.*', 'COMMENT'),
    (r'(\s|\\\n|\\\r\n)+', 'SPACE'),
    (r'["\'](.|\n|\r)*["\']', 'STRING'),
    (r'[uU]["\'](.|\n|\r)*["\']', 'UNICODE_STRING'),
    (r'[fF]["\'](.|\n|\r)*["\']', 'INTERPOLATED_STRING'),
    (r'[rR]["\'](.|\n|\r)*["\']', 'RAW_STRING'),
    (r'[bB]["\'](.|\n|\r)*["\']', 'BINARY_STRING'),
    (r'[uU][rR]["\'](.|\n|\r)*["\']', 'UNICODE_RAW_STRING'),
    (r'[bB][rR]["\'](.|\n|\r)*["\']', 'BINARY_RAW_STRING'),
    (r'[fF][rR]["\'](.|\n|\r)*["\']', 'INTERPOLATED_RAW_STRING'),
    (r'[rR][fF]["\'](.|\n|\r)*["\']', 'INTERPOLATED_RAW_STRING'),
)


TOKENS = [(re.compile('^' + x[0] + '$'), x[1]) for x in TOKENS]


def tokenize(sequence, print_function=False):
    return list(tokenize_generator(sequence, print_function))


def tokenize_current_keywords(print_function=False):
    if print_function is True:
        return [x for x in KEYWORDS if x not in ("print", "exec")]
    else:
        return KEYWORDS


def tokenize_generator(sequence, print_function=False):
    current_keywords = tokenize_current_keywords()

    for item in sequence:
        if item in current_keywords:
            yield (item.upper(), item)
            continue

        for candidate, token_name in TOKENS:
            if candidate.match(item):
                yield (token_name, item)
                break
        else:
            raise UnknowItem("Can't find a matching token for this item: '%s'" % item)
    yield ('ENDMARKER', '')
    yield



#!/usr/bin/python
# -*- coding:Utf-8 -*-
from .test_utils import parse_simple


def test_return():
    "return"
    parse_simple([
        ('RETURN', 'return'),
    ], [{"type": "return", "value": None, "formatting": [], }])


def test_return_a():
    "return a"
    parse_simple([
        ('RETURN', 'return', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
    ], [{"type": "return", "value": {"type": "name", "value": 'a', }, "formatting": [{"type": "space", "value": " "}], }])


def test_yield():
    "yield"
    parse_simple([
        ('YIELD', 'yield'),
    ], [{
        "type": "yield",
        "value": None,
        "formatting": [],
    }])


def test_yield_from():
    "yield from a"
    parse_simple([
        ('YIELD', 'yield', [], [('SPACE', ' ')]),
        ('FROM', 'from', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
    ], [{
        "type": "yield_from",
        "first_formatting": [{"type": "space", "value": " "}],
        "value": {"type": "name", "value": 'a', },
        "formatting": [{"type": "space", "value": " "}], }])


def test_yield_a():
    "yield a"
    parse_simple([
        ('YIELD', 'yield', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
    ], [{"type": "yield", "value": {"type": "name", "value": 'a', }, "formatting": [{"type": "space", "value": " "}], }])


def test_del():
    "del a"
    parse_simple([
        ('DEL', 'del', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
    ], [
        {
            "type": "del",
            "value": {"type": "name", "value": 'a', },
            "formatting": [{"type": "space", "value": " "}],
        }
    ])


def test_break():
    "break"
    parse_simple([
        ('BREAK', 'break'),
    ], [
        {
            "type": "break",
        }
    ])


def test_continue():
    "continue"
    parse_simple([
        ('CONTINUE', 'continue'),
    ], [
        {
            "type": "continue",
        }
    ])


def test_pass():
    "pass"
    parse_simple([
        ('PASS', 'pass'),
    ], [
        {
            "type": "pass",
        }
    ])


def test_assert():
    "assert a"
    parse_simple([
        ('ASSERT', 'assert', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
    ], [
        {
            "type": "assert",
            "value": {"type": "name", "value": 'a', },
            "message": None,
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [],
            "third_formatting": []
        }
    ])


def test_assert_message():
    "assert a , b"
    parse_simple([
        ('ASSERT', 'assert', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('COMMA', ',', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'b'),
    ], [
        {
            "type": "assert",
            "value": {"type": "name", "value": 'a', },
            "message": {"type": "name", "value": 'b'},
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [{"type": "space", "value": " "}],
            "third_formatting": [{"type": "space", "value": " "}]
        }
    ])
    parse_simple([
        ('ASSERT', 'assert', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('COMMA', ',', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'b'),
    ], [
        {
            "type": "assert",
            "value": {"type": "name", "value": 'a'},
            "message": {"type": "name", "value": 'b', },
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [{"type": "space", "value": " "}],
            "third_formatting": [{"type": "space", "value": " "}]
        }
    ])


def test_raise_empty():
    "raise"
    parse_simple([
        ('RAISE', 'raise', [], [('SPACE', ' ')]),
    ], [
        {
            "type": "raise",
            "value": None,
            "instance": None,
            "traceback": None,
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [],
            "third_formatting": [],
            "fourth_formatting": [],
            "fifth_formatting": [],
            "comma_or_from": None,
        }
    ])


def test_raise():
    "raise a"
    parse_simple([
        ('RAISE', 'raise', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
    ], [
        {
            "type": "raise",
            "value": {"type": "name", "value": 'a'},
            "instance": None,
            "traceback": None,
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [],
            "third_formatting": [],
            "fourth_formatting": [],
            "fifth_formatting": [],
            "comma_or_from": None,
        }
    ])


def test_raise_from():
    "raise a from b"
    parse_simple([
        ('RAISE', 'raise', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('FROM', 'from', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'b'),
    ], [
        {
            "type": "raise",
            "comma_or_from": "from",
            "value": {"type": "name", "value": 'a'},
            "instance": {'type': 'name', 'value': 'b'},
            "traceback": None,
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [{'type': 'space', 'value': ' '}],
            "third_formatting": [{'type': 'space', 'value': ' '}],
            "fourth_formatting": [],
            "fifth_formatting": [],
        }
    ])


def test_raise_instance():
    "raise a, b"
    parse_simple([
        ('RAISE', 'raise', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('COMMA', 'comma', [], [('SPACE', ' ')]),
        ('NAME', 'b'),
    ], [
        {
            "type": "raise",
            "comma_or_from": ",",
            "value": {"type": "name", "value": 'a'},
            "instance": {"type": "name", "value": 'b'},
            "traceback": None,
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [],
            "third_formatting": [{"type": "space", "value": " "}],
            "fourth_formatting": [],
            "fifth_formatting": [],
        }
    ])
    parse_simple([
        ('RAISE', 'raise', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('COMMA', 'comma', [], [('SPACE', ' ')]),
        ('NAME', 'b'),
    ], [
        {
            "type": "raise",
            "comma_or_from": ",",
            "value": {"type": "name", "value": 'a'},
            "instance": {"type": "name", "value": 'b'},
            "traceback": None,
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [],
            "third_formatting": [{"type": "space", "value": " "}],
            "fourth_formatting": [],
            "fifth_formatting": [],
        }
    ])


def test_raise_instance_traceback():
    "raise a, b, c"
    parse_simple([
        ('RAISE', 'raise', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('COMMA', 'comma', [], [('SPACE', ' ')]),
        ('NAME', 'b'),
        ('COMMA', 'comma', [], [('SPACE', ' ')]),
        ('NAME', 'c'),
    ], [
        {
            "type": "raise",
            "comma_or_from": ",",
            "value": {"type": "name", "value": 'a'},
            "instance": {"type": "name", "value": 'b'},
            "traceback": {"type": "name", "value": 'c'},
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [],
            "third_formatting": [{"type": "space", "value": " "}],
            "fourth_formatting": [],
            "fifth_formatting": [{"type": "space", "value": " "}],
        }
    ])
    parse_simple([
        ('RAISE', 'raise', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('COMMA', 'comma', [], [('SPACE', ' ')]),
        ('NAME', 'b'),
        ('COMMA', 'comma', [], [('SPACE', ' ')]),
        ('NAME', 'c'),
    ], [
        {
            "type": "raise",
            "comma_or_from": ",",
            "value": {"type": "name", "value": 'a'},
            "instance": {"type": "name", "value": 'b'},
            "traceback": {"type": "name", "value": 'c'},
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [],
            "third_formatting": [{"type": "space", "value": " "}],
            "fourth_formatting": [],
            "fifth_formatting": [{"type": "space", "value": " "}],
        }
    ])
    parse_simple([
        ('RAISE', 'raise', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('COMMA', 'comma', [], [('SPACE', ' ')]),
        ('NAME', 'b'),
        ('COMMA', 'comma', [], [('SPACE', ' ')]),
        ('NAME', 'c'),
    ], [
        {
            "type": "raise",
            "comma_or_from": ",",
            "value": {"type": "name", "value": 'a'},
            "instance": {"type": "name", "value": 'b'},
            "traceback": {"type": "name", "value": 'c'},
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [],
            "third_formatting": [{"type": "space", "value": " "}],
            "fourth_formatting": [],
            "fifth_formatting": [{"type": "space", "value": " "}],
        }
    ])


def test_exec():
    "exec a"
    parse_simple([
        ('EXEC', 'exec', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
    ], [
        {
            "type": "exec",
            "value": {"type": "name", "value": 'a', },
            "globals": None,
            "locals": None,
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [],
            "third_formatting": [],
            "fourth_formatting": [],
            "fifth_formatting": []
        }
    ])


def test_exec_in():
    "exec a in b"
    parse_simple([
        ('EXEC', 'exec', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('IN', 'in', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'b'),
    ], [
        {
            "type": "exec",
            "value": {"type": "name", "value": 'a', },
            "globals": {"type": "name", "value": 'b'},
            "locals": None,
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [{"type": "space", "value": " "}],
            "third_formatting": [{"type": "space", "value": " "}],
            "fourth_formatting": [],
            "fifth_formatting": []
        }
    ])
    parse_simple([
        ('EXEC', 'exec', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('IN', 'in', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'b'),
    ], [
        {
            "type": "exec",
            "value": {"type": "name", "value": 'a'},
            "globals": {"type": "name", "value": 'b', },
            "locals": None,
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [{"type": "space", "value": " "}],
            "third_formatting": [{"type": "space", "value": " "}],
            "fourth_formatting": [],
            "fifth_formatting": []
        }
    ])


def test_exec_in_c():
    "exec a in b, c"
    parse_simple([
        ('EXEC', 'exec', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('IN', 'in', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'b'),
        ('COMMA', ',', [], [('SPACE', ' ')]),
        ('NAME', 'c'),
    ], [
        {
            "type": "exec",
            "value": {"type": "name", "value": 'a', },
            "globals": {"type": "name", "value": 'b'},
            "locals": {"type": "name", "value": 'c'},
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [{"type": "space", "value": " "}],
            "third_formatting": [{"type": "space", "value": " "}],
            "fourth_formatting": [],
            "fifth_formatting": [{"type": "space", "value": " "}]
        }
    ])
    parse_simple([
        ('EXEC', 'exec', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('IN', 'in', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'b'),
        ('COMMA', ',', [], [('SPACE', ' ')]),
        ('NAME', 'c'),
    ], [
        {
            "type": "exec",
            "value": {"type": "name", "value": 'a'},
            "globals": {"type": "name", "value": 'b', },
            "locals": {"type": "name", "value": 'c'},
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [{"type": "space", "value": " "}],
            "third_formatting": [{"type": "space", "value": " "}],
            "fourth_formatting": [],
            "fifth_formatting": [{"type": "space", "value": " "}]
        }
    ])
    parse_simple([
        ('EXEC', 'exec', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('IN', 'in', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'b'),
        ('COMMA', ',', [], [('SPACE', ' ')]),
        ('NAME', 'c'),
    ], [
        {
            "type": "exec",
            "value": {"type": "name", "value": 'a'},
            "globals": {"type": "name", "value": 'b'},
            "locals": {"type": "name", "value": 'c', },
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [{"type": "space", "value": " "}],
            "third_formatting": [{"type": "space", "value": " "}],
            "fourth_formatting": [],
            "fifth_formatting": [{"type": "space", "value": " "}]
        }
    ])


def test_global():
    "global a"
    parse_simple([
        ('GLOBAL', 'global', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
    ], [
        {
            "type": "global",
            "formatting": [{"type": "space", "value": " "}],
            "value": [
                {"type": "name", "value": 'a', },
            ]
        }
    ])


def test_global_one():
    "global a, b"
    parse_simple([
        ('GLOBAL', 'global', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('COMMA', ',', [], [('SPACE', ' ')]),
        ('NAME', 'b'),
    ], [
        {
            "type": "global",
            "formatting": [{"type": "space", "value": " "}],
            "value": [
                {"type": "name", "value": 'a', },
                {
                    "type": "comma",
                    "first_formatting": [],
                    "second_formatting": [{"type": "space", "value": " ", }, ],
                },
                {"type": "name", "value": 'b'},
            ]
        }
    ])
    parse_simple([
        ('GLOBAL', 'global', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('COMMA', ',', [], [('SPACE', ' ')]),
        ('NAME', 'b'),
    ], [
        {
            "type": "global",
            "formatting": [{"type": "space", "value": " "}],
            "value": [
                {"type": "name", "value": 'a'},
                {
                    "type": "comma",
                    "first_formatting": [],
                    "second_formatting": [{"type": "space", "value": " ", }, ],
                },
                {"type": "name", "value": 'b', },
            ]
        }
    ])


def test_global_two():
    "global a, b ,  c"
    parse_simple([
        ('GLOBAL', 'global', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('COMMA', ',', [], [('SPACE', ' ')]),
        ('NAME', 'b'),
        ('COMMA', ',', [('SPACE', ' ')], [('SPACE', '  ')]),
        ('NAME', 'c'),
    ], [
        {
            "type": "global",
            "formatting": [{"type": "space", "value": " "}],
            "value": [
                {"type": "name", "value": "a"},
                {
                    "type": "comma",
                    "first_formatting": [],
                    "second_formatting": [{"type": "space", "value": " ", }, ],
                },
                {"type": "name", "value": "b"},
                {
                    "type": "comma",
                    "first_formatting": [{"type": "space", "value": " "}, ],
                    "second_formatting": [{"type": "space", "value": "  "}, ],
                },
                {"type": "name", "value": "c"},
            ]
        }
    ])


def test_print():
    "print"
    parse_simple([
        ('PRINT', 'print', [], []),
    ], [
        {
            "type": "print",
            "formatting": [],
            "value": [],
            "destination_formatting": [],
            "destination": None,
        }
    ])


def test_print_a():
    "print a"
    parse_simple([
        ('PRINT', 'print', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
    ], [
        {
            "type": "print",
            "formatting": [{"type": "space", "value": " "}],
            "value": [{"type": "name", "value": 'a', }],
            "destination_formatting": [],
            "destination": None,
        }
    ])


def test_print_a_b():
    "print a, b"
    parse_simple([
        ('PRINT', 'print', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('COMMA', ',', [], [('SPACE', ' ')]),
        ('NAME', 'b'),
    ], [
        {
            "type": "print",
            "formatting": [
                {
                    "type": "space",
                    "value": " "
                }
            ],
            "value": [
                {
                    "type": "name",
                    "value": 'a',
                },
                {
                    "type": "comma",
                    "first_formatting": [],
                    "second_formatting": [{"type": "space", "value": " "}]
                },
                {
                    "type": "name",
                    "value": 'b'
                },
            ],
            "destination_formatting": [],
            "destination": None,
        }
    ])
    parse_simple([
        ('PRINT', 'print', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('COMMA', ',', [], [('SPACE', ' ')]),
        ('NAME', 'b'),
    ], [
        {
            "type": "print",
            "formatting": [
                {
                    "type": "space",
                    "value": " "
                }
            ],
            "value": [
                {
                    "type": "name",
                    "value": 'a'
                },
                {
                    "type": "comma",
                    "first_formatting": [],
                    "second_formatting": [{"type": "space", "value": " "}]
                },
                {
                    "type": "name",
                    "value": 'b',
                },
            ],
            "destination_formatting": [],
            "destination": None,
        }
    ])


def test_print_a_b_comma():
    "print a, b,"
    parse_simple([
        ('PRINT', 'print', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('COMMA', ',', [], [('SPACE', ' ')]),
        ('NAME', 'b'),
        ('COMMA', ',', [], []),
    ], [
        {
            "type": "print",
            "formatting": [
                {
                    "type": "space",
                    "value": " "
                }
            ],
            "value": [
                {
                    "type": "name",
                    "value": 'a',
                },
                {
                    "type": "comma",
                    "first_formatting": [],
                    "second_formatting": [{"type": "space", "value": " "}]
                },
                {
                    "type": "name",
                    "value": 'b'
                },
                {
                    "type": "comma",
                    "first_formatting": [],
                    "second_formatting": []
                },
            ],
            "destination_formatting": [],
            "destination": None,
        }
    ])
    parse_simple([
        ('PRINT', 'print', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('COMMA', ',', [], [('SPACE', ' ')]),
        ('NAME', 'b'),
        ('COMMA', ',', [], []),
    ], [
        {
            "type": "print",
            "formatting": [
                {
                    "type": "space",
                    "value": " "
                }
            ],
            "value": [
                {
                    "type": "name",
                    "value": 'a'
                },
                {
                    "type": "comma",
                    "first_formatting": [],
                    "second_formatting": [{"type": "space", "value": " "}]
                },
                {
                    "type": "name",
                    "value": 'b',
                },
                {
                    "type": "comma",
                    "first_formatting": [],
                    "second_formatting": []
                },
            ],
            "destination_formatting": [],
            "destination": None,
        }
    ])


def test_print_redirect():
    "print >> a"
    parse_simple([
        ('PRINT', 'print', [], [('SPACE', ' ')]),
        ('RIGHT_SHIFT', '>>', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
    ], [
        {
            "type": "print",
            "formatting": [
                {
                    "type": "space",
                    "value": " "
                }
            ],
            "value": [],
            "destination_formatting": [
                {
                    "type": "space",
                    "value": " "
                }
            ],
            "destination": {
                "type": "name",
                "value": 'a',
            },
        }
    ])


def test_print_redirect_ab():
    "print >> a , b"
    parse_simple([
        ('PRINT', 'print', [], [('SPACE', ' ')]),
        ('RIGHT_SHIFT', '>>', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('COMMA', ',', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'b'),
    ], [
        {
            "type": "print",
            "formatting": [
                {
                    "type": "space",
                    "value": " "
                }
            ],
            "value": [
                {
                    "type": "comma",
                    "first_formatting": [{"type": "space", "value": " "}],
                    "second_formatting": [{"type": "space", "value": " "}]
                },
                {
                    "type": "name",
                    "value": 'b',
                },
            ],
            "destination": {"type": "name", "value": 'a'},
            "destination_formatting": [{"type": "space", "value": " "}],
        }
    ])
    parse_simple([
        ('PRINT', 'print', [], [('SPACE', ' ')]),
        ('RIGHT_SHIFT', '>>', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('COMMA', ',', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'b'),
    ], [
        {
            "type": "print",
            "formatting": [{"type": "space", "value": " "}],
            "value": [
                {
                    "type": "comma",
                    "first_formatting": [{"type": "space", "value": " "}],
                    "second_formatting": [{"type": "space", "value": " "}]
                },
                {
                    "type": "name",
                    "value": 'b'
                },
            ],
            "destination": {
                "type": "name",
                "value": 'a',
            },
            "destination_formatting": [
                {
                    "type": "space",
                    "value": " "
                }
            ],
        }
    ])


def test_print_redirect_ab_comma():
    "print >> a , b ,"
    parse_simple([
        ('PRINT', 'print', [], [('SPACE', ' ')]),
        ('RIGHT_SHIFT', '>>', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('COMMA', ',', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'b'),
        ('COMMA', ',', [('SPACE', ' ')]),
    ], [
        {
            "type": "print",
            "formatting": [
                {
                    "type": "space",
                    "value": " "
                }
            ],
            "value": [
                {
                    "type": "comma",
                    "first_formatting": [{"type": "space", "value": " "}],
                    "second_formatting": [{"type": "space", "value": " "}]
                },
                {
                    "type": "name",
                    "value": 'b',
                },
                {
                    "type": "comma",
                    "first_formatting": [{"type": "space", "value": " "}],
                    "second_formatting": []
                },
            ],
            "destination": {"type": "name", "value": 'a'},
            "destination_formatting": [{"type": "space", "value": " "}],
        }
    ])
    parse_simple([
        ('PRINT', 'print', [], [('SPACE', ' ')]),
        ('RIGHT_SHIFT', '>>', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('COMMA', ',', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'b'),
        ('COMMA', ',', [('SPACE', ' ')]),
    ], [
        {
            "type": "print",
            "formatting": [
                {
                    "type": "space",
                    "value": " "
                }
            ],
            "value": [
                {
                    "type": "comma",
                    "first_formatting": [{"type": "space", "value": " "}],
                    "second_formatting": [{"type": "space", "value": " "}]
                },
                {
                    "type": "name",
                    "value": 'b'
                },
                {
                    "type": "comma",
                    "first_formatting": [{"type": "space", "value": " "}],
                    "second_formatting": []
                },
            ],
            "destination": {
                "type": "name",
                "value": 'a',
            },
            "destination_formatting": [
                {
                    "type": "space",
                    "value": " "
                }
            ],
        }
    ])


def test_lambda():
    "lambda : a"
    parse_simple([
        ('LAMBDA', 'lambda', [], [('SPACE', ' ')]),
        ('COLON', ':', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
    ], [
        {
            "type": "lambda",
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [],
            "third_formatting": [{"type": "space", "value": " "}],
            "value": {
                "type": "name",
                "value": "a",
            },
            "arguments": []
        }
    ])


def test_lambda_arguments():
    "lambda argument : a"
    parse_simple([
        ('LAMBDA', 'lambda', [], [('SPACE', ' ')]),
        ('NAME', 'argument'),
        ('COLON', ':', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'a'),
    ], [
        {
            "type": "lambda",
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [{"type": "space", "value": " "}],
            "third_formatting": [{"type": "space", "value": " "}],
            "value": {
                "type": "name",
                "value": "a",
            },
            "arguments": [
                {
                    "value": {},
                    "first_formatting": [],
                    "second_formatting": [],
                    "type": "def_argument",
                    "annotation": {},
                    "annotation_first_formatting": [],
                    "annotation_second_formatting": [],
                    "target": {
                        "value": "argument",
                        "type": "name",
                    }
                }
            ]
        }
    ])



#!/usr/bin/python
# -*- coding:Utf-8 -*-


from baron.spliter import split, UntreatedError
from baron.utils import python_version
import pytest


def test_empty():
    assert split("") == []


def test_print():
    assert split("print") == ["print"]


def test_print_space():
    assert split("print ") == ['print', ' ']


def test_import():
    assert split("import  pouet") == ["import", "  ", "pouet"]


def test_from_import():
    assert split(" from zob   import  pouet ") == [" ", "from", " ", "zob", "   ", "import", "  ", "pouet", " "]


def test_from_import_as():
    assert split("from a import b as c") == ["from", " ", "a", " ", "import", " ", "b", " ", "as", " ", "c"]


def test_underscore_variable():
    assert split("some_variable") == ["some_variable"]


def test_different_case():
    assert split("AbCd cDeF") == ["AbCd", " ", "cDeF"]


def test_decorator():
    assert split("@pouet") == ["@", "pouet"]


def test_tab_n_space():
    assert split("	 ") == ["	 "]


def test_several_spaces():
    assert split("     ") == ["     "]


def test_numbers():
    assert split("1234") == ["1234"]


def test_several_numbers():
    assert split("12 34") == ["12", " ", "34"]


def test_comma():
    assert split(",") == [","]


def test_comma_with_words():
    assert split("a, b") == ["a", ",", " ", "b"]


def test_dot():
    assert split(".") == ["."]


def test_dot_with_word():
    assert split("a.b") == ["a", ".", "b"]


def test_dot_with_words():
    assert split("a.b.c") == ["a", ".", "b", ".", "c"]


def test_colon():
    assert split(";") == [";"]


def test_colon_word():
    assert split("pouet;") == ["pouet", ";"]


def test_assign():
    assert split("a = b") == ["a", " ", "=", " ", "b"]


def test_call():
    assert split("function()") == ["function", "(", ")"]


def test_call_with_arg():
    assert split("function(a)") == ["function", "(", "a", ")"]


def test_call_with_args():
    assert split("function(a, b, c)") == ["function", "(", "a", ",", " ", "b", ",", " ", "c", ")"]


def test_call_with_kwarg():
    assert split("function(a=b)") == ["function", "(", "a", "=", "b", ")"]


def test_call_with_kwargs():
    assert split("function(a=b, c= d)") == ["function", "(", "a", "=", "b", ",", " ", "c", "=", " ", "d", ")"]


def test_call_with_start_args():
    assert split("function(*args)") == ["function", "(", "*", "args", ")"]


def test_call_with_start_kwargs():
    assert split("function(**kwargs)") == ["function", "(", "*", "*", "kwargs", ")"]


def test_function_def():
    assert split("def pouet(): pass") == ["def", " ", "pouet", "(", ")", ":", " ", "pass"]


def test_addition():
    assert split("a + 2") == ["a", " ", "+", " ", "2"]


def test_substraction():
    assert split("a - 2") == ["a", " ", "-", " ", "2"]


def test_multiplication():
    assert split("a * 2") == ["a", " ", "*", " ", "2"]


def test_division():
    assert split("a/2") == ["a", "/", "2"]


def test_power():
    assert split("a**2") == ["a", "*", "*", "2"]


def test_modulo():
    assert split("a % 2") == ["a", " ", "%", " ", "2"]


def test_binary_stuff():
    assert split("a^2") == ["a", "^", "2"]
    assert split("a&2") == ["a", "&", "2"]
    assert split("a|2") == ["a", "|", "2"]
    assert split("a << 2") == ["a", " ", "<", "<", " ", "2"]
    assert split("a >> 2") == ["a", " ", ">", ">", " ", "2"]


def test_operator_assign():
    assert split("a += 2") == ["a", " ", "+", "=", " ", "2"]
    assert split("a -= 2") == ["a", " ", "-", "=", " ", "2"]
    assert split("a *= 2") == ["a", " ", "*", "=", " ", "2"]
    assert split("a /= 2") == ["a", " ", "/", "=", " ", "2"]
    assert split("a %= 2") == ["a", " ", "%", "=", " ", "2"]
    assert split("a &= 2") == ["a", " ", "&", "=", " ", "2"]
    assert split("a |= 2") == ["a", " ", "|", "=", " ", "2"]
    assert split("a ^= 2") == ["a", " ", "^", "=", " ", "2"]
    assert split("a //= 2") == ["a", " ", "/", "/", "=", " ", "2"]
    assert split("a **= 2") == ["a", " ", "*", "*", "=", " ", "2"]
    assert split("a <<= 2") == ["a", " ", "<", "<", "=", " ", "2"]
    assert split("a >>= 2") == ["a", " ", ">", ">", "=", " ", "2"]


def test_factor():
    assert split("~a") == ["~", "a"]


def test_del_pouet():
    assert split("del pouet") == ["del", " ", "pouet"]


def test_pass():
    assert split("pass") == ["pass"]


def test_break():
    assert split("break") == ["break"]


def test_continue():
    assert split("continue") == ["continue"]


def test_return():
    assert split("return") == ["return"]


def test_return_stuff():
    assert split("return stuff") == ["return", " ", "stuff"]


def test_yield():
    assert split("yield") == ["yield"]


def test_yield_stuff():
    assert split("yield stuff") == ["yield", " ", "stuff"]


def test_raise():
    assert split("raise") == ["raise"]


def test_raise_stuff():
    assert split("raise Exception()") == ["raise", " ", "Exception", "(", ")"]


def test_global_stuff():
    assert split("global stuff") == ["global", " ", "stuff"]


def test_exec():
    assert split("exec") == ["exec"]


def test_exec_stuff():
    assert split("exec stuff") == ["exec", " ", "stuff"]


def test_assert():
    assert split("assert") == ["assert"]


def test_assert_stuff():
    assert split("assert stuff") == ["assert", " ", "stuff"]


def test_line_end():
    assert split("\n") == ["\n"]


def test_line_end_windows():
    assert split("\r\n") == ["\r", "\n"]


def test_if():
    assert split("if ab: pass") == ["if", " ", "ab", ":", " ", "pass"]


def test_if_elif_else():
    assert split("if a:\n pass\nelif b:\n pass\nelse: \n pass") == ["if", " ", "a", ":", "\n", " ", "pass", "\n", "elif", " ", "b", ":", "\n", " ", "pass", "\n", "else", ":", " ", "\n", " ", "pass"]


def test_while():
    assert split("while a: pass") == ["while", " ", "a", ":", " ", "pass"]


def test_lambda():
    assert split("lambda x: x + 1") == ["lambda", " ", "x", ":", " ", "x", " ", "+", " ", "1"]


def test_for():
    assert split("for a in b: pass") == ["for", " ", "a", " ", "in", " ", "b", ":", " ", "pass"]


def test_empty_list():
    assert split("[]") == ["[", "]"]


def test_list():
    assert split("[a, b, c]") == ["[", "a", ",", " ", "b", ",", " ", "c", "]"]


def test_empty_dict():
    assert split("{}") == ["{", "}"]


def test_dict():
    assert split("{a: b, c: d}") == ["{", "a", ":", " ", "b", ",", " ", "c", ":", " ", "d", "}"]


def test_not():
    assert split("!a") == ["!", "a"]


def test_not_equal():
    assert split("a != b") == ["a", " ", "!", "=", " ", "b"]


def test_backquote():
    assert split("`a`") == ["`", "a", "`"]


def test_number_in_var():
    assert split("a1") == ["a1"]


def test_comment():
    assert split("# a b c d") == ["# a b c d"]


def test_comments():
    assert split("# a b c d\n# pouet") == ["# a b c d", "\n", "# pouet"]


def test_empty_string():
    assert split("''") == ["''"]


def test_string():
    assert split("'pouet pouet'") == ["'pouet pouet'"]


def test_empty_string_other_quotes():
    assert split('"pouet pouet"') == ['"pouet pouet"']


def test_multi_string():
    assert split("'''pouet pouet'''") == ["'''pouet pouet'''"]


def test_multi_string_other_quotes():
    assert split('"""pouet pouet"""') == ['"""pouet pouet"""']


def test_missing_quote_yields_error():
    with pytest.raises(UntreatedError):
        split("'")

    with pytest.raises(UntreatedError):
        split("'''")

    with pytest.raises(UntreatedError):
        split('"')

    with pytest.raises(UntreatedError):
        split('"""')


def test_escape():
    assert split("\\\\") == ["\\", "\\"]


def test_escape_in_string():
    assert split("'\\\\'") == ["'\\\\'"]


def test_other_escape_string():
    assert split("'\\\\'") == ["'\\\\'"]


def test_hexa():
    assert split("0x7F") == ["0x7F"]


def test_multi_string_with_same_quotes_in():
    assert split('"""pouet " "" pouet"""') == ['"""pouet " "" pouet"""']


def test_comment_backslash():
    assert split('# pouet \\\npouet') == ["# pouet \\", "\n", "pouet"]


def test_backslash_in_comment():
    assert split("# pouet \\t pouet\npouet") == ["# pouet \\t pouet", "\n", "pouet"]


def test_regression():
    assert split("(r'[\"\\'](.|\n|\r)*[\"\\']', 'STRING'),") == ["(", "r", "'[\"\\'](.|\n|\r)*[\"\\']'", ",", " ", "'STRING'", ")", ","]


# TODO: make this test pass in python3 also
# requires to remove dependency on ast.py
if python_version == 2:
    def test_remove_crap():
        assert split("\x0c\xef\xbb\xbf") == []


def test_split_float_notation():
    assert split("a._") == ["a", ".", "_"]



#!/usr/bin/python
# -*- coding:Utf-8 -*-

import pytest
from baron.parser import ParsingError
from .test_utils import parse_multi


def test_if_stmt():
    "if a: pass"
    parse_multi([
        ('IF', 'if', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('COLON', ':'),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
    ], [
        {
            "type": "ifelseblock",
            "value": [
                {
                    "type": "if",
                    "first_formatting": [{"type": "space", "value": " "}],
                    "second_formatting": [],
                    "third_formatting": [],
                    "test": {
                        "type": "name",
                        "value": "a",
                    },
                    "value": [
                        {
                            "type": "pass",
                        },
                        {
                            "formatting": [],
                            "indent": "",
                            "type": "endl",
                            "value": "\n"
                        }
                    ],
                }
            ],
        }
    ])


def test_if_stmt_indent():
    """
    if a:
        pass
    """
    parse_multi([
        ('IF', 'if', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('COLON', ':'),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
        ('DEDENT', ''),
    ], [
        {
            "type": "ifelseblock",
            "value": [
                {
                    "type": "if",
                    "first_formatting": [{"type": "space", "value": " "}],
                    "second_formatting": [],
                    "third_formatting": [],
                    "test": {
                        "type": "name",
                        "value": "a",
                    },
                    "value": [
                        {
                            "formatting": [],
                            "type": "endl",
                            "value": "\n",
                            "indent": "    "
                        },
                        {
                            "type": "pass",
                        },
                        {
                            "formatting": [],
                            "indent": "",
                            "type": "endl",
                            "value": "\n"
                        }
                    ],
                }
            ]
        }
    ])


def test_if_stmt_indent_two_endls():
    """
    if a:

        pass
    """
    parse_multi([
        ('IF', 'if', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('COLON', ':'),
        ('ENDL', '\n'),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
        ('DEDENT', ''),
    ], [
        {
            "type": "ifelseblock",
            "value": [
                {
                    "type": "if",
                    "first_formatting": [{"type": "space", "value": " "}],
                    "second_formatting": [],
                    "third_formatting": [],
                    "test": {
                        "type": "name",
                        "value": "a",
                    },
                    "value": [
                        {
                            "type": "endl",
                            "value": "\n",
                            "formatting": [],
                            "indent": ""
                        },
                        {
                            "type": "endl",
                            "value": "\n",
                            "formatting": [],
                            "indent": "    "
                        },
                        {
                            "type": "pass",
                        },
                        {
                            "formatting": [],
                            "indent": "",
                            "type": "endl",
                            "value": "\n"
                        }
                    ],
                }
            ]
        }
    ])


def test_if_stmt_indent_multiple_endls():
    """
    if a:


        pass
    """
    parse_multi([
        ('IF', 'if', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('COLON', ':'),
        ('ENDL', '\n'),
        ('ENDL', '\n', [], [('SPACE', '  ')]),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
        ('DEDENT', ''),
    ], [
        {
            "type": "ifelseblock",
            "value": [
                {
                    "type": "if",
                    "first_formatting": [{"type": "space", "value": " "}],
                    "second_formatting": [],
                    "third_formatting": [],
                    "test": {
                        "type": "name",
                        "value": "a",
                    },
                    "value": [
                        {
                            "type": "endl",
                            "formatting": [],
                            "value": "\n",
                            "indent": ""
                        },
                        {
                            "type": "endl",
                            "formatting": [],
                            "value": "\n",
                            "indent": "  "
                        },
                        {
                            "type": "endl",
                            "formatting": [],
                            "value": "\n",
                            "indent": "    "
                        },
                        {
                            "type": "pass",
                        },
                        {
                            "formatting": [],
                            "indent": "",
                            "type": "endl",
                            "value": "\n"
                        }
                    ],
                }
            ]
        }
    ])


def test_if_else_stmt_indent():
    """
    if a:
        pass
    else:
        pass
    """
    parse_multi([
        ('IF', 'if', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('COLON', ':'),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
        ('DEDENT', ''),
        ('ELSE', 'else', [], [('SPACE', ' ')]),
        ('COLON', ':'),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
        ('DEDENT', ''),
    ], [
        {
            "type": "ifelseblock",
            "value": [
                {
                    "type": "if",
                    "first_formatting": [{"type": "space", "value": " "}],
                    "second_formatting": [],
                    "third_formatting": [],
                    "test": {
                        "type": "name",
                        "value": "a",
                    },
                    "value": [
                        {
                            "type": "endl",
                            "value": "\n",
                            "formatting": [],
                            "indent": "    "
                        },
                        {
                            "type": "pass",
                        },
                        {
                            "formatting": [],
                            "indent": "",
                            "type": "endl",
                            "value": "\n"
                        }
                    ],
                },
                {
                    "type": "else",
                    "first_formatting": [{"type": "space", "value": " "}],
                    "second_formatting": [],
                    "value": [
                        {
                            "type": "endl",
                            "value": "\n",
                            "formatting": [],
                            "indent": "    "
                        },
                        {
                            "type": "pass",
                        },
                        {
                            "formatting": [],
                            "indent": "",
                            "type": "endl",
                            "value": "\n",
                        }
                    ]
                }
            ]
        }
    ])


def test_if_elif_elif_stmt_indent():
    """
    if a:
        pass
    elif b:
        pass
    elif c :
        pass
    """
    parse_multi([
        ('IF', 'if', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('COLON', ':'),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
        ('DEDENT', ''),
        ('ELIF', 'elif', [], [('SPACE', ' ')]),
        ('NAME', 'b'),
        ('COLON', ':'),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
        ('DEDENT', ''),
        ('ELIF', 'elif', [], [('SPACE', ' ')]),
        ('NAME', 'c'),
        ('COLON', ':', [('SPACE', ' ')]),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
        ('DEDENT', ''),
    ], [
        {
            "type": "ifelseblock",
            "value": [
                {
                    "type": "if",
                    "first_formatting": [{"type": "space", "value": " "}],
                    "second_formatting": [],
                    "third_formatting": [],
                    "test": {
                        "type": "name",
                        "value": "a",
                    },
                    "value": [
                        {
                            "type": "endl",
                            "value": "\n",
                            "formatting": [],
                            "indent": "    "
                        },
                        {
                            "type": "pass",
                        },
                        {
                            "formatting": [],
                            "indent": "",
                            "type": "endl",
                            "value": "\n"
                        }
                    ],
                },
                {
                    "type": "elif",
                    "first_formatting": [{"type": "space", "value": " "}],
                    "second_formatting": [],
                    "third_formatting": [],
                    "test": {
                        "type": "name",
                        "value": "b",
                    },
                    "value": [
                        {
                            "type": "endl",
                            "value": "\n",
                            "formatting": [],
                            "indent": "    "
                        },
                        {
                            "type": "pass",
                        },
                        {
                            "formatting": [],
                            "indent": "",
                            "type": "endl",
                            "value": "\n"
                        }
                    ],
                },
                {
                    "type": "elif",
                    "first_formatting": [{"type": "space", "value": " "}],
                    "second_formatting": [{"type": "space", "value": " "}],
                    "third_formatting": [],
                    "test": {
                        "type": "name",
                        "value": "c",
                    },
                    "value": [
                        {
                            "type": "endl",
                            "formatting": [],
                            "value": "\n",
                            "indent": "    "
                        },
                        {
                            "type": "pass",
                        },
                        {
                            "formatting": [],
                            "indent": "",
                            "type": "endl",
                            "value": "\n"
                        }
                    ],
                }
            ]
        }
    ])


def test_if_elif_elif_else_stmt_indent():
    """
    if a:
        pass
    elif b:
        pass
    elif c :
        pass
    else:
        pass
    """
    parse_multi([
        ('IF', 'if', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('COLON', ':'),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
        ('DEDENT', ''),
        ('ELIF', 'elif', [], [('SPACE', ' ')]),
        ('NAME', 'b'),
        ('COLON', ':'),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
        ('DEDENT', ''),
        ('ELIF', 'elif', [], [('SPACE', ' ')]),
        ('NAME', 'c'),
        ('COLON', ':', [('SPACE', ' ')]),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
        ('DEDENT', ''),
        ('ELSE', 'else', [], [('SPACE', ' ')]),
        ('COLON', ':'),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
        ('DEDENT', ''),
    ], [
        {
            "type": "ifelseblock",
            "value": [
                {
                    "type": "if",
                    "first_formatting": [{"type": "space", "value": " "}],
                    "second_formatting": [],
                    "third_formatting": [],
                    "test": {
                        "type": "name",
                        "value": "a",
                    },
                    "value": [
                        {
                            "type": "endl",
                            "formatting": [],
                            "value": "\n",
                            "indent": "    "
                        },
                        {
                            "type": "pass",
                        },
                        {
                            "formatting": [],
                            "indent": "",
                            "type": "endl",
                            "value": "\n"
                        }
                    ],
                },
                {
                    "type": "elif",
                    "first_formatting": [{"type": "space", "value": " "}],
                    "second_formatting": [],
                    "third_formatting": [],
                    "test": {
                        "type": "name",
                        "value": "b",
                    },
                    "value": [
                        {
                            "type": "endl",
                            "value": "\n",
                            "formatting": [],
                            "indent": "    "
                        },
                        {
                            "type": "pass",
                        },
                        {
                            "type": "endl",
                            "value": "\n",
                            "formatting": [],
                            "indent": "",
                        }
                    ],
                },
                {
                    "type": "elif",
                    "first_formatting": [{"type": "space", "value": " "}],
                    "second_formatting": [{"type": "space", "value": " "}],
                    "third_formatting": [],
                    "test": {
                        "type": "name",
                        "value": "c",
                    },
                    "value": [
                        {
                            "type": "endl",
                            "formatting": [],
                            "value": "\n",
                            "indent": "    "
                        },
                        {
                            "type": "pass",
                        },
                        {
                            "formatting": [],
                            "indent": "",
                            "type": "endl",
                            "value": "\n"
                        }
                    ],
                },
                {
                    "type": "else",
                    "first_formatting": [{"type": "space", "value": " "}],
                    "second_formatting": [],
                    "value": [
                        {
                            "formatting": [],
                            "type": "endl",
                            "value": "\n",
                            "indent": "    "
                        },
                        {
                            "type": "pass",
                        },
                        {
                            "formatting": [],
                            "indent": "",
                            "type": "endl",
                            "value": "\n",
                        }
                    ]
                }
            ]
        }
    ])


def test_while_stmt_indent():
    """
    while a:
        pass
    """
    parse_multi([
        ('WHILE', 'while', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('COLON', ':'),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
        ('DEDENT', ''),
    ], [
        {
            "type": "while",
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [],
            "third_formatting": [],
            "else": {},
            "test": {
                "type": "name",
                "value": "a",
            },
            "value": [
                {
                    "type": "endl",
                    "formatting": [],
                    "value": "\n",
                    "indent": "    "
                },
                {
                    "type": "pass",
                },
                {
                    "formatting": [],
                    "indent": "",
                    "type": "endl",
                    "value": "\n"
                }
            ],
        }
    ])


def test_while_stmt_indent_third_formatting():
    """
    while a :
        pass
    """
    parse_multi([
        ('WHILE', 'while', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('COLON', ':', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
        ('DEDENT', ''),
    ], [
        {
            "type": "while",
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [{"type": "space", "value": " "}],
            "third_formatting": [{"type": "space", "value": " "}],
            "else": {},
            "test": {
                "type": "name",
                "value": "a",
            },
            "value": [
                {
                    "type": "endl",
                    "formatting": [],
                    "value": "\n",
                    "indent": "    "
                },
                {
                    "type": "pass",
                },
                {
                    "formatting": [],
                    "indent": "",
                    "type": "endl",
                    "value": "\n"
                }
            ],
        }
    ])


def test_while_else_stmt_indent():
    """
    while a:
        pass
    else:
        pass
    """
    parse_multi([
        ('WHILE', 'while', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('COLON', ':'),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
        ('DEDENT', ''),
        ('ELSE', 'else', [], [('SPACE', " ")]),
        ('COLON', ':'),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
        ('DEDENT', ''),
    ], [
        {
            "type": "while",
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [],
            "third_formatting": [],
            "test": {
                "type": "name",
                "value": "a",
            },
            "value": [
                {
                    "type": "endl",
                    "value": "\n",
                    "formatting": [],
                    "indent": "    "
                },
                {
                    "type": "pass",
                },
                {
                    "formatting": [],
                    "indent": "",
                    "type": "endl",
                    "value": "\n"
                }
            ],
            "else": {
                "type": "else",
                "first_formatting": [{"type": "space", "value": " "}],
                "second_formatting": [],
                "value": [
                    {
                        "type": "endl",
                        "value": "\n",
                        "formatting": [],
                        "indent": "    "
                    },
                    {
                        "type": "pass",
                    },
                    {
                        "indent": "",
                        "formatting": [],
                        "type": "endl",
                        "value": "\n",
                    }
                ]
            }
        }
    ])


def test_for_stmt_indent():
    """
    for i in a:
        pass
    """
    parse_multi([
        ('FOR', 'for', [], [('SPACE', ' ')]),
        ('NAME', 'i'),
        ('IN', 'in', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('COLON', ':', [], [('SPACE', ' ')]),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
        ('DEDENT', ''),
    ], [
        {
            "type": "for",
            "async": False,
            "async_formatting": [],
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [{"type": "space", "value": " "}],
            "third_formatting": [{"type": "space", "value": " "}],
            "fourth_formatting": [],
            "fifth_formatting": [{"type": "space", "value": " "}],
            "else": {},
            "iterator": {
                "type": "name",
                "value": "i",
            },
            "target": {
                "type": "name",
                "value": "a",
            },
            "value": [
                {
                    "type": "endl",
                    "value": "\n",
                    "formatting": [],
                    "indent": "    "
                },
                {
                    "type": "pass",
                },
                {
                    "indent": "",
                    "formatting": [],
                    "type": "endl",
                    "value": "\n"
                }
            ],
        }
    ])


def test_async_for_stmt_indent():
    """
    async for i in a:
        pass
    """
    parse_multi([
        ('NAME', 'async', [], []),
        ('SPACE', ' '),
        ('FOR', 'for', [], [('SPACE', ' ')]),
        ('NAME', 'i'),
        ('IN', 'in', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('COLON', ':', [], [('SPACE', ' ')]),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
        ('DEDENT', ''),
    ], [
        {
            "type": "for",
            "async": True,
            "async_formatting": [{"type": "space", "value": " "}],
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [{"type": "space", "value": " "}],
            "third_formatting": [{"type": "space", "value": " "}],
            "fourth_formatting": [],
            "fifth_formatting": [{"type": "space", "value": " "}],
            "else": {},
            "iterator": {
                "type": "name",
                "value": "i",
            },
            "target": {
                "type": "name",
                "value": "a",
            },
            "value": [
                {
                    "type": "endl",
                    "value": "\n",
                    "formatting": [],
                    "indent": "    "
                },
                {
                    "type": "pass",
                },
                {
                    "indent": "",
                    "formatting": [],
                    "type": "endl",
                    "value": "\n"
                }
            ],
        }
    ])


def test_async_for_stmt_indent_bad_keyword():
    """
    not_async for i in a:
        pass
    """
    with pytest.raises(ParsingError):
        parse_multi([
            ('NAME', 'not_async', [], []),
            ('SPACE', ' '),
            ('FOR', 'for', [], [('SPACE', ' ')]),
            ('NAME', 'i'),
            ('IN', 'in', [('SPACE', ' ')], [('SPACE', ' ')]),
            ('NAME', 'a'),
            ('COLON', ':', [], [('SPACE', ' ')]),
            ('ENDL', '\n', [], [('SPACE', '    ')]),
            ('INDENT', ''),
            ('PASS', 'pass'),
            ('ENDL', '\n'),
            ('DEDENT', ''),
        ], [])


def test_for_else_stmt_indent():
    """
    for i in b:
        pass
    else:
        pass
    """
    parse_multi([
        ('FOR', 'for', [], [('SPACE', ' ')]),
        ('NAME', 'i'),
        ('IN', 'in', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'b'),
        ('COLON', ':'),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
        ('DEDENT', ''),
        ('ELSE', 'else'),
        ('COLON', ':'),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
        ('DEDENT', ''),
    ], [
        {
            "type": "for",
            "async": False,
            "async_formatting": [],
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [{"type": "space", "value": " "}],
            "third_formatting": [{"type": "space", "value": " "}],
            "fourth_formatting": [],
            "fifth_formatting": [],
            "else": {
                "type": "else",
                "first_formatting": [],
                "second_formatting": [],
                "value": [
                    {
                        "type": "endl",
                        "value": "\n",
                        "formatting": [],
                        "indent": "    "
                    },
                    {
                        "type": "pass",
                    },
                    {
                        "indent": "",
                        "formatting": [],
                        "type": "endl",
                        "value": "\n",
                    }
                ]
            },
            "iterator": {
                "type": "name",
                "value": "i",
            },
            "target": {
                "type": "name",
                "value": "b",
            },
            "value": [
                {
                    "type": "endl",
                    "formatting": [],
                    "value": "\n",
                    "indent": "    "
                },
                {
                    "type": "pass",
                },
                {
                    "indent": "",
                    "formatting": [],
                    "type": "endl",
                    "value": "\n"
                }
            ],
        }
    ])


def test_async_for_else_stmt_indent():
    """
    async for i in b:
        pass
    else:
        pass
    """
    parse_multi([
        ('NAME', 'async', [], []),
        ('SPACE', ' '),
        ('FOR', 'for', [], [('SPACE', ' ')]),
        ('NAME', 'i'),
        ('IN', 'in', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'b'),
        ('COLON', ':'),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
        ('DEDENT', ''),
        ('ELSE', 'else'),
        ('COLON', ':'),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
        ('DEDENT', ''),
    ], [
        {
            "type": "for",
            "async": True,
            "async_formatting": [{"type": "space", "value": " "}],
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [{"type": "space", "value": " "}],
            "third_formatting": [{"type": "space", "value": " "}],
            "fourth_formatting": [],
            "fifth_formatting": [],
            "else": {
                "type": "else",
                "first_formatting": [],
                "second_formatting": [],
                "value": [
                    {
                        "type": "endl",
                        "value": "\n",
                        "formatting": [],
                        "indent": "    "
                    },
                    {
                        "type": "pass",
                    },
                    {
                        "indent": "",
                        "formatting": [],
                        "type": "endl",
                        "value": "\n",
                    }
                ]
            },
            "iterator": {
                "type": "name",
                "value": "i",
            },
            "target": {
                "type": "name",
                "value": "b",
            },
            "value": [
                {
                    "type": "endl",
                    "formatting": [],
                    "value": "\n",
                    "indent": "    "
                },
                {
                    "type": "pass",
                },
                {
                    "indent": "",
                    "formatting": [],
                    "type": "endl",
                    "value": "\n"
                }
            ],
        }
    ])


def test_async_for_else_stmt_indent_bad_keyword():
    """
    async for i in b:
        pass
    else:
        pass
    """
    with pytest.raises(ParsingError):
        parse_multi([
            ('NAME', 'not_async', [], []),
            ('SPACE', ' '),
            ('FOR', 'for', [], [('SPACE', ' ')]),
            ('NAME', 'i'),
            ('IN', 'in', [('SPACE', ' ')], [('SPACE', ' ')]),
            ('NAME', 'b'),
            ('COLON', ':'),
            ('ENDL', '\n', [], [('SPACE', '    ')]),
            ('INDENT', ''),
            ('PASS', 'pass'),
            ('ENDL', '\n'),
            ('DEDENT', ''),
            ('ELSE', 'else'),
            ('COLON', ':'),
            ('ENDL', '\n', [], [('SPACE', '    ')]),
            ('INDENT', ''),
            ('PASS', 'pass'),
            ('ENDL', '\n'),
            ('DEDENT', ''),
        ], [])


def test_try_finally_stmt_indent():
    """
    try :
        pass
    finally :
        pass
    """
    parse_multi([
        ('TRY', 'try'),
        ('COLON', ':', [('SPACE', ' ')]),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
        ('DEDENT', ''),
        ('FINALLY', 'finally'),
        ('COLON', ':', [('SPACE', ' ')]),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
        ('DEDENT', ''),
    ], [
        {
            "type": "try",
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [],
            "else": {},
            "finally": {
                "type": "finally",
                "first_formatting": [{"type": "space", "value": " "}],
                "second_formatting": [],
                "value": [
                    {
                        "type": "endl",
                        "value": "\n",
                        "formatting": [],
                        "indent": "    "
                    },
                    {
                        "type": "pass",
                    },
                    {
                        "indent": "",
                        "formatting": [],
                        "type": "endl",
                        "value": "\n"
                    }
                ],
            },
            "excepts": [],
            "value": [
                {
                    "type": "endl",
                    "value": "\n",
                    "formatting": [],
                    "indent": "    "
                },
                {
                    "type": "pass",
                },
                {
                    "indent": "",
                    "formatting": [],
                    "type": "endl",
                    "value": "\n"
                }
            ],
        }
    ])


def test_try_excepts_stmt_empty():
    """
    try :
        pass
    except:
        pass
    """
    parse_multi([
        ('TRY', 'try'),
        ('COLON', ':', [('SPACE', ' ')]),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
        ('DEDENT', ''),
        ('EXCEPT', 'except', [], []),
        ('COLON', ':'),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
        ('DEDENT', ''),
    ], [
        {
            "type": "try",
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [],
            "else": {},
            "finally": {},
            "excepts": [
                {
                    "type": "except",
                    "first_formatting": [],
                    "second_formatting": [],
                    "third_formatting": [],
                    "fourth_formatting": [],
                    "fifth_formatting": [],
                    "delimiter": "",
                    "target": {},
                    "exception": {},
                    "value": [
                        {
                            "type": "endl",
                            "formatting": [],
                            "value": "\n",
                            "indent": "    "
                        },
                        {
                            "type": "pass",
                        },
                        {
                            "indent": "",
                            "formatting": [],
                            "type": "endl",
                            "value": "\n"
                        }
                    ]
                }
            ],
            "value": [
                {
                    "type": "endl",
                    "formatting": [],
                    "value": "\n",
                    "indent": "    "
                },
                {
                    "type": "pass",
                },
                {
                    "indent": "",
                    "formatting": [],
                    "type": "endl",
                    "value": "\n"
                }
            ],
        }
    ])


def test_try_excepts_stmt_indent():
    """
    try :
        pass
    except IOError:
        pass
    except Exception:
        pass
    """
    parse_multi([
        ('TRY', 'try'),
        ('COLON', ':', [('SPACE', ' ')]),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
        ('DEDENT', ''),
        ('EXCEPT', 'except', [], [('SPACE', ' ')]),
        ('NAME', 'IOError'),
        ('COLON', ':'),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
        ('DEDENT', ''),
        ('EXCEPT', 'except', [], [('SPACE', ' ')]),
        ('NAME', 'Exception'),
        ('COLON', ':'),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
        ('DEDENT', ''),
    ], [
        {
            "type": "try",
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [],
            "else": {},
            "finally": {},
            "excepts": [
                {
                    "type": "except",
                    "first_formatting": [{"type": "space", "value": " "}],
                    "second_formatting": [],
                    "third_formatting": [],
                    "fourth_formatting": [],
                    "fifth_formatting": [],
                    "delimiter": "",
                    "target": {},
                    "exception": {
                        "type": "name",
                        "value": "IOError",
                    },
                    "value": [
                        {
                            "type": "endl",
                            "formatting": [],
                            "value": "\n",
                            "indent": "    "
                        },
                        {
                            "type": "pass",
                        },
                        {
                            "indent": "",
                            "formatting": [],
                            "type": "endl",
                            "value": "\n"
                        }
                    ]
                },
                {
                    "type": "except",
                    "first_formatting": [{"type": "space", "value": " "}],
                    "second_formatting": [],
                    "third_formatting": [],
                    "fourth_formatting": [],
                    "fifth_formatting": [],
                    "delimiter": "",
                    "target": {},
                    "exception": {
                        "type": "name",
                        "value": "Exception",
                    },
                    "value": [
                        {
                            "type": "endl",
                            "value": "\n",
                            "formatting": [],
                            "indent": "    "
                        },
                        {
                            "type": "pass",
                        },
                        {
                            "indent": "",
                            "formatting": [],
                            "type": "endl",
                            "value": "\n"
                        }
                    ]
                }
            ],
            "value": [
                {
                    "type": "endl",
                    "formatting": [],
                    "value": "\n",
                    "indent": "    "
                },
                {
                    "type": "pass",
                },
                {
                    "indent": "",
                    "formatting": [],
                    "type": "endl",
                    "value": "\n"
                }
            ],
        }
    ])


def test_try_except_comma_stmt_indent():
    """
    try :
        pass
    except IOError, e:
        pass
    """
    parse_multi([
        ('TRY', 'try'),
        ('COLON', ':', [('SPACE', ' ')]),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
        ('DEDENT', ''),
        ('EXCEPT', 'except', [], [('SPACE', ' ')]),
        ('NAME', 'IOError'),
        ('COMMA', ',', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('COLON', ':'),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
        ('DEDENT', ''),
    ], [
        {
            "type": "try",
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [],
            "else": {},
            "finally": {},
            "excepts": [
                {
                    "type": "except",
                    "first_formatting": [{"type": "space", "value": " "}],
                    "second_formatting": [],
                    "third_formatting": [{"type": "space", "value": " "}],
                    "fourth_formatting": [],
                    "fifth_formatting": [],
                    "delimiter": ",",
                    "target": {
                        "type": "name",
                        "value": "a"
                    },
                    "exception": {
                        "type": "name",
                        "value": "IOError",
                    },
                    "value": [
                        {
                            "type": "endl",
                            "value": "\n",
                            "formatting": [],
                            "indent": "    "
                        },
                        {
                            "type": "pass",
                        },
                        {
                            "indent": "",
                            "formatting": [],
                            "type": "endl",
                            "value": "\n"
                        }
                    ]
                }
            ],
            "value": [
                {
                    "formatting": [],
                    "type": "endl",
                    "value": "\n",
                    "indent": "    "
                },
                {
                    "type": "pass",
                },
                {
                    "formatting": [],
                    "indent": "",
                    "type": "endl",
                    "value": "\n"
                }
            ],
        }
    ])


def test_try_except_comma_stmt_else_indent():
    """
    try :
        pass
    except IOError, e:
        pass
    else:
        pass
    """
    parse_multi([
        ('TRY', 'try'),
        ('COLON', ':', [('SPACE', ' ')]),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
        ('DEDENT', ''),
        ('EXCEPT', 'except', [], [('SPACE', ' ')]),
        ('NAME', 'IOError'),
        ('COMMA', ',', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('COLON', ':'),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
        ('DEDENT', ''),
        ('ELSE', 'else'),
        ('COLON', ':'),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
        ('DEDENT', ''),
    ], [
        {
            "type": "try",
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [],
            "else": {
                "type": "else",
                "first_formatting": [],
                "second_formatting": [],
                "value": [
                    {
                        "type": "endl",
                        "value": "\n",
                        "formatting": [],
                        "indent": "    "
                    },
                    {
                        "type": "pass",
                    },
                    {
                        "type": "endl",
                        "value": "\n",
                        "formatting": [],
                        "indent": ""
                    }
                ]
            },
            "finally": {},
            "excepts": [
                {
                    "type": "except",
                    "first_formatting": [{"type": "space", "value": " "}],
                    "second_formatting": [],
                    "third_formatting": [{"type": "space", "value": " "}],
                    "fourth_formatting": [],
                    "fifth_formatting": [],
                    "delimiter": ",",
                    "target": {
                        "type": "name",
                        "value": "a"
                    },
                    "exception": {
                        "type": "name",
                        "value": "IOError",
                    },
                    "value": [
                        {
                            "type": "endl",
                            "formatting": [],
                            "value": "\n",
                            "indent": "    "
                        },
                        {
                            "type": "pass",
                        },
                        {
                            "formatting": [],
                            "indent": "",
                            "type": "endl",
                            "value": "\n"
                        }
                    ]
                }
            ],
            "value": [
                {
                    "type": "endl",
                    "value": "\n",
                    "formatting": [],
                    "indent": "    "
                },
                {
                    "type": "pass",
                },
                {
                    "formatting": [],
                    "indent": "",
                    "type": "endl",
                    "value": "\n"
                }
            ],
        }
    ])


def test_try_except_comma_stmt_else_finally_indent():
    """
    try :
        pass
    except IOError, e:
        pass
    else:
        pass
    finally:
        pass
    """
    parse_multi([
        ('TRY', 'try'),
        ('COLON', ':', [('SPACE', ' ')]),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
        ('DEDENT', ''),
        ('EXCEPT', 'except', [], [('SPACE', ' ')]),
        ('NAME', 'IOError'),
        ('COMMA', ',', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('COLON', ':'),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
        ('DEDENT', ''),
        ('ELSE', 'else'),
        ('COLON', ':'),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
        ('DEDENT', ''),
        ('FINALLY', 'finally'),
        ('COLON', ':'),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
        ('DEDENT', ''),
    ], [
        {
            "type": "try",
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [],
            "else": {
                "type": "else",
                "first_formatting": [],
                "second_formatting": [],
                "value": [
                    {
                        "type": "endl",
                        "value": "\n",
                        "formatting": [],
                        "indent": "    "
                    },
                    {
                        "type": "pass",
                    },
                    {
                        "type": "endl",
                        "value": "\n",
                        "formatting": [],
                        "indent": ""
                    }
                ]
            },
            "finally": {
                "type": "finally",
                "first_formatting": [],
                "second_formatting": [],
                "value": [
                    {
                        "type": "endl",
                        "value": "\n",
                        "formatting": [],
                        "indent": "    "
                    },
                    {
                        "type": "pass",
                    },
                    {
                        "type": "endl",
                        "value": "\n",
                        "formatting": [],
                        "indent": "",
                    }
                ],
            },
            "excepts": [
                {
                    "type": "except",
                    "first_formatting": [{"type": "space", "value": " "}],
                    "second_formatting": [],
                    "third_formatting": [{"type": "space", "value": " "}],
                    "fourth_formatting": [],
                    "fifth_formatting": [],
                    "delimiter": ",",
                    "target": {
                        "type": "name",
                        "value": "a"
                    },
                    "exception": {
                        "type": "name",
                        "value": "IOError",
                    },
                    "value": [
                        {
                            "type": "endl",
                            "formatting": [],
                            "value": "\n",
                            "indent": "    "
                        },
                        {
                            "type": "pass",
                        },
                        {
                            "formatting": [],
                            "indent": "",
                            "type": "endl",
                            "value": "\n"
                        }
                    ]
                }
            ],
            "value": [
                {
                    "type": "endl",
                    "value": "\n",
                    "formatting": [],
                    "indent": "    "
                },
                {
                    "type": "pass",
                },
                {
                    "formatting": [],
                    "indent": "",
                    "type": "endl",
                    "value": "\n"
                }
            ],
        }
    ])


def test_try_except_comma_stmt_finally_indent():
    """
    try :
        pass
    except IOError, e:
        pass
    finally:
        pass
    """
    parse_multi([
        ('TRY', 'try'),
        ('COLON', ':', [('SPACE', ' ')]),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
        ('DEDENT', ''),
        ('EXCEPT', 'except', [], [('SPACE', ' ')]),
        ('NAME', 'IOError'),
        ('COMMA', ',', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('COLON', ':'),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
        ('DEDENT', ''),
        ('FINALLY', 'finally'),
        ('COLON', ':'),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
        ('DEDENT', ''),
    ], [
        {
            "type": "try",
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [],
            "else": {},
            "finally": {
                "type": "finally",
                "first_formatting": [],
                "second_formatting": [],
                "value": [
                    {
                        "type": "endl",
                        "value": "\n",
                        "formatting": [],
                        "indent": "    "
                    },
                    {
                        "type": "pass",
                    },
                    {
                        "formatting": [],
                        "indent": "",
                        "type": "endl",
                        "value": "\n"
                    }
                ],
            },
            "excepts": [
                {
                    "type": "except",
                    "first_formatting": [{"type": "space", "value": " "}],
                    "second_formatting": [],
                    "third_formatting": [{"type": "space", "value": " "}],
                    "fourth_formatting": [],
                    "fifth_formatting": [],
                    "delimiter": ",",
                    "target": {
                        "type": "name",
                        "value": "a"
                    },
                    "exception": {
                        "type": "name",
                        "value": "IOError",
                    },
                    "value": [
                        {
                            "type": "endl",
                            "formatting": [],
                            "value": "\n",
                            "indent": "    "
                        },
                        {
                            "type": "pass",
                        },
                        {
                            "formatting": [],
                            "indent": "",
                            "type": "endl",
                            "value": "\n"
                        }
                    ]
                }
            ],
            "value": [
                {
                    "type": "endl",
                    "formatting": [],
                    "value": "\n",
                    "indent": "    "
                },
                {
                    "type": "pass",
                },
                {
                    "formatting": [],
                    "indent": "",
                    "type": "endl",
                    "value": "\n"
                }
            ],
        }
    ])


def test_try_except_as_stmt_indent():
    """
    try :
        pass
    except IOError as e:
        pass
    """
    parse_multi([
        ('TRY', 'try'),
        ('COLON', ':', [('SPACE', ' ')]),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
        ('DEDENT', ''),
        ('EXCEPT', 'except', [], [('SPACE', ' ')]),
        ('NAME', 'IOError'),
        ('AS', 'as', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('COLON', ':'),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
        ('DEDENT', ''),
    ], [
        {
            "type": "try",
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [],
            "else": {},
            "finally": {},
            "excepts": [
                {
                    "type": "except",
                    "first_formatting": [{"type": "space", "value": " "}],
                    "second_formatting": [{"type": "space", "value": " "}],
                    "third_formatting": [{"type": "space", "value": " "}],
                    "fourth_formatting": [],
                    "fifth_formatting": [],
                    "delimiter": "as",
                    "target": {
                        "type": "name",
                        "value": "a"
                    },
                    "exception": {
                        "type": "name",
                        "value": "IOError",
                    },
                    "value": [
                        {
                            "type": "endl",
                            "value": "\n",
                            "formatting": [],
                            "indent": "    "
                        },
                        {
                            "type": "pass",
                        },
                        {
                            "formatting": [],
                            "indent": "",
                            "type": "endl",
                            "value": "\n"
                        }
                    ]
                }
            ],
            "value": [
                {
                    "type": "endl",
                    "value": "\n",
                    "formatting": [],
                    "indent": "    "
                },
                {
                    "type": "pass",
                },
                {
                    "formatting": [],
                    "indent": "",
                    "type": "endl",
                    "value": "\n"
                }
            ],
        }
    ])



from .token import BaronToken
from .parser import BaronParserGenerator

from .tokenizer import TOKENS, tokenize, tokenize_current_keywords
from .utils import create_node_from_token
from .grammator_imports import include_imports
from .grammator_control_structures import include_control_structures
from .grammator_primitives import include_primivites
from .grammator_operators import include_operators
from .grammator_data_structures import include_data_structures
from .parser import ParsingError


def generate_parse(print_function):
    pg = BaronParserGenerator(tuple([x.upper() for x in tokenize_current_keywords(print_function)] + [x[1] for x in TOKENS] + ["ENDMARKER", "INDENT", "DEDENT"]), cache_id="baron")

    @pg.production("main : statements")
    def main(pack):
        (statements,) = pack
        return [x for x in statements if x] if statements else []

    @pg.production("statements : statements statement")
    def statements_statement(pack):
        (statements, statement) = pack
        return statements + statement

    @pg.production("statements : statement SEMICOLON")
    def statement_semicolon(pack):
        (statement, semicolon) = pack
        return statement +\
            [{
                "type": "semicolon",
                "first_formatting": semicolon.hidden_tokens_before,
                "second_formatting": semicolon.hidden_tokens_after,
                "value": ";"
            }]

    @pg.production("statements : statement")
    def statement(pack):
        (statement,) = pack
        return statement

    @pg.production("statement : endl")
    def statement_endl(pack):
        (endl,) = pack
        return endl

    @pg.production("endl : ENDL")
    def endl(pack):
        (endl,) = pack
        indent = ""
        if endl.hidden_tokens_after and endl.hidden_tokens_after[0]["type"] == "space":
            indent = endl.hidden_tokens_after[0]["value"]
            endl.hidden_tokens_after = endl.hidden_tokens_after[1:]
        return [{
            "type": "endl",
            "value": endl.value,
            "formatting": endl.hidden_tokens_before,
            "indent": indent,
        }] + endl.hidden_tokens_after

    @pg.production("left_parenthesis : LEFT_PARENTHESIS")
    def left_parenthesis(pack):
        (lp,) = pack
        return lp

    @pg.production("endl : COMMENT ENDL")
    def comment(pack):
        (comment_, endl) = pack
        return [{
            "type": "comment",
            "value": comment_.value,
            "formatting": comment_.hidden_tokens_before,
        }, {
            "type": "endl",
            "formatting": endl.hidden_tokens_before,
            "indent": endl.hidden_tokens_after[0]["value"] if endl.hidden_tokens_after else "",
            "value": endl.value
        }]

    @pg.production("statement : ENDMARKER")
    def end(_):
        return [None]

    @pg.production("statement : simple_stmt")
    @pg.production("statement : compound_stmt")
    def statement_simple_statement(pack):
        (stmt,) = pack
        return stmt

    @pg.production("simple_stmt : small_stmt SEMICOLON endl")
    def simple_stmt_semicolon_endl(pack):
        (small_stmt, semicolon, endl) = pack
        return [small_stmt,
                {
                    "type": "semicolon",
                    "value": ";",
                    "first_formatting": semicolon.hidden_tokens_before,
                    "second_formatting": semicolon.hidden_tokens_after
                }] + endl

    @pg.production("simple_stmt : small_stmt endl")
    def simple_stmt(pack):
        (small_stmt, endl) = pack
        return [small_stmt] + endl

    @pg.production("simple_stmt : small_stmt SEMICOLON simple_stmt")
    def simple_stmt_semicolon(pack):
        (small_stmt, semicolon, simple_stmt) = pack
        return [small_stmt,
                {
                    "type": "semicolon",
                    "value": ";",
                    "first_formatting": semicolon.hidden_tokens_before,
                    "second_formatting": semicolon.hidden_tokens_after
                }] + simple_stmt

    @pg.production("small_stmt : flow_stmt")
    @pg.production("small_stmt : del_stmt")
    @pg.production("small_stmt : pass_stmt")
    @pg.production("small_stmt : assert_stmt")
    @pg.production("small_stmt : raise_stmt")
    @pg.production("small_stmt : global_stmt")
    @pg.production("small_stmt : nonlocal_stmt")
    @pg.production("compound_stmt : if_stmt")
    @pg.production("compound_stmt : while_stmt")
    @pg.production("compound_stmt : for_stmt")
    @pg.production("compound_stmt : try_stmt")
    @pg.production("compound_stmt : funcdef")
    @pg.production("compound_stmt : classdef")
    @pg.production("compound_stmt : with_stmt")
    @pg.production("compound_stmt : decorated")
    @pg.production("compound_stmt : async_stmt")
    def small_and_compound_stmt(pack):
        (statement,) = pack
        return statement

    @pg.production("async_maybe : ")
    def async_maybe(pack):
        return {
            "async": False,
            "formatting": [],
        }

    @pg.production("async_maybe : NAME")
    @pg.production("async : NAME")
    def async_without_space(pack):
        (async_,) = pack
        return {
            "async": True,
            "value": async_.value,
            "formatting": [],
        }

    @pg.production("async_maybe : NAME SPACE")
    @pg.production("async : NAME SPACE")
    def async_(pack):
        (async_, space) = pack
        return {
            "async": True,
            "value": async_.value,
            "formatting": [{'type': 'space', 'value': space.value}],
        }

    @pg.production("async_stmt : async with_stmt")
    @pg.production("async_stmt : async for_stmt")
    def async_stmt(pack):
        (async_, statement,) = pack

        if async_["value"] != "async":
            raise ParsingError("The only possible keyword before a '%s' is 'async', not '%s'" % (statement[0]["type"], async_["value"]))

        statement[0]["async"] = True
        statement[0]["async_formatting"] += async_["formatting"]
        return statement

    if not print_function:
        @pg.production("small_stmt : print_stmt")
        def print_statement(pack):
            (statement,) = pack
            return statement

    @pg.production("small_stmt : expr_stmt")
    @pg.production("expr_stmt : testlist_star_expr")
    @pg.production("testlist : test")
    @pg.production("testlist_argslist : test")
    @pg.production("testlist_star_expr : test_or_star_expr")
    @pg.production("test : or_test")
    @pg.production("test : lambdef")
    @pg.production("or_test : and_test")
    @pg.production("and_test : not_test")
    @pg.production("not_test : comparison")
    @pg.production("comparison : expr")
    @pg.production("expr : xor_expr")
    @pg.production("xor_expr : and_expr")
    @pg.production("and_expr : shift_expr")
    @pg.production("shift_expr : arith_expr")
    @pg.production("arith_expr : term")
    @pg.production("term : factor")
    @pg.production("factor : power")
    @pg.production("power : atom")
    @pg.production("exprlist : expr")
    def term_factor(pack):
        (level,) = pack
        return level

    @pg.production("with_stmt : WITH with_items COLON suite")
    def with_stmt(pack):
        (with_, with_items, colon, suite) = pack
        return [{
            "type": "with",
            "async": False,
            "async_formatting": [],
            "value": suite,
            "first_formatting": with_.hidden_tokens_after,
            "second_formatting": colon.hidden_tokens_before,
            "third_formatting": colon.hidden_tokens_after,
            "contexts": with_items
        }]

    @pg.production("with_items : with_items comma with_item")
    def with_items_with_item(pack):
        (with_items, comma, with_item,) = pack
        return with_items + [comma, with_item]

    @pg.production("with_items : with_item")
    def with_items(pack):
        (with_item,) = pack
        return [with_item]

    @pg.production("with_item : test")
    def with_item(pack):
        (test,) = pack
        return {
            "type": "with_context_item",
            "as": {},
            "first_formatting": [],
            "second_formatting": [],
            "value": test
        }

    @pg.production("with_item : test AS expr")
    def with_item_as(pack):
        (test, as_, expr) = pack
        return {
            "type": "with_context_item",
            "as": expr,
            "first_formatting": as_.hidden_tokens_before,
            "second_formatting": as_.hidden_tokens_after,
            "value": test
        }

    @pg.production("classdef : CLASS NAME COLON suite")
    def class_stmt(pack,):
        (class_, name, colon, suite) = pack
        return [{
            "type": "class",
            "name": name.value,
            "parenthesis": False,
            "first_formatting": class_.hidden_tokens_after,
            "second_formatting": [],
            "third_formatting": [],
            "fourth_formatting": [],
            "fifth_formatting": colon.hidden_tokens_before,
            "sixth_formatting": colon.hidden_tokens_after,
            "inherit_from": [],
            "decorators": [],
            "value": suite,
        }]

    @pg.production("classdef : CLASS NAME LEFT_PARENTHESIS RIGHT_PARENTHESIS COLON suite")
    def class_stmt_parenthesis(pack,):
        (class_, name, left_parenthesis, right_parenthesis, colon, suite) = pack
        return [{
            "type": "class",
            "name": name.value,
            "parenthesis": True,
            "first_formatting": class_.hidden_tokens_after,
            "second_formatting": left_parenthesis.hidden_tokens_before,
            "third_formatting": left_parenthesis.hidden_tokens_after,
            "fourth_formatting": right_parenthesis.hidden_tokens_before,
            "fifth_formatting": right_parenthesis.hidden_tokens_after + colon.hidden_tokens_before,
            "sixth_formatting": colon.hidden_tokens_after,
            "inherit_from": [],
            "decorators": [],
            "value": suite,
        }]

    @pg.production("classdef : CLASS NAME LEFT_PARENTHESIS testlist_argslist RIGHT_PARENTHESIS COLON suite")
    def class_stmt_inherit(pack,):
        def unfold_simple_call_arguments(node):
            if node.get("type") == "call_argument" and not node["target"]:
                return node["value"]
            return node

        (class_, name, left_parenthesis, testlist, right_parenthesis, colon, suite) = pack
        return [{
            "type": "class",
            "name": name.value,
            "parenthesis": True,
            "first_formatting": class_.hidden_tokens_after,
            "second_formatting": left_parenthesis.hidden_tokens_before,
            "third_formatting": left_parenthesis.hidden_tokens_after,
            "fourth_formatting": right_parenthesis.hidden_tokens_before,
            "fifth_formatting": right_parenthesis.hidden_tokens_after + colon.hidden_tokens_before,
            "sixth_formatting": colon.hidden_tokens_after,
            "inherit_from": [unfold_simple_call_arguments(x)
                             for x in (testlist if isinstance(testlist, list) else [testlist])],
            "decorators": [],
            "value": suite,
        }]

    @pg.production("decorated : decorators funcdef")
    @pg.production("decorated : decorators classdef")
    def decorated(pack):
        (decorators, funcdef) = pack
        funcdef[0]["decorators"] = decorators
        return funcdef

    @pg.production("decorators : decorators decorator")
    def decorators_decorator(pack):
        (decorators, decorator,) = pack
        return decorators + decorator

    @pg.production("decorators : decorator")
    def decorators(pack):
        (decorator,) = pack
        return decorator

    # TODO tests
    @pg.production("decorator : endl")
    def decorator_endl(pack):
        # thanks iPython devs, you appear to be the only one in the world that
        # split decorators with empty lines... like seriously.
        (endl,) = pack
        return endl

    @pg.production("decorator : AT dotted_name endl")
    def decorator(pack):
        (at, dotted_name, endl) = pack
        return [{
            "type": "decorator",
            "value": {
                "value": dotted_name,
                "type": "dotted_name",
            },
            "call": {},
            "formatting": at.hidden_tokens_after,
        }] + endl

    @pg.production("decorator : AT dotted_name LEFT_PARENTHESIS RIGHT_PARENTHESIS endl")
    def decorator_empty_call(pack):
        (at, dotted_name, left_parenthesis, right_parenthesis, endl) = pack
        return [{
            "type": "decorator",
            "value": {
                "value": dotted_name,
                "type": "dotted_name",
            },
            "call": {
                "third_formatting": right_parenthesis.hidden_tokens_before,
                "fourth_formatting": right_parenthesis.hidden_tokens_after,
                "type": "call",
                "first_formatting": left_parenthesis.hidden_tokens_before,
                "value": [],
                "second_formatting": left_parenthesis.hidden_tokens_after
            },
            "formatting": at.hidden_tokens_after,
        }] + endl

    @pg.production("decorator : AT dotted_name LEFT_PARENTHESIS argslist RIGHT_PARENTHESIS endl")
    def decorator_call(pack):
        (at, dotted_name, left_parenthesis, argslist, right_parenthesis, endl) = pack
        return [{
            "type": "decorator",
            "value": {
                "value": dotted_name,
                "type": "dotted_name",
            },
            "call": {
                "third_formatting": right_parenthesis.hidden_tokens_before,
                "fourth_formatting": right_parenthesis.hidden_tokens_after,
                "type": "call",
                "first_formatting": left_parenthesis.hidden_tokens_before,
                "value": argslist,
                "second_formatting": left_parenthesis.hidden_tokens_after
            },
            "formatting": at.hidden_tokens_after,
        }] + endl

    @pg.production("funcdef : async_maybe DEF NAME LEFT_PARENTHESIS typed_parameters RIGHT_PARENTHESIS return_annotation COLON suite")
    def function_definition(pack):
        (async_maybe, def_, name, left_parenthesis, typed_parameters, right_parenthesis, return_annotation, colon, suite) = pack

        if async_maybe["async"] and async_maybe["value"] != "async":
            raise ParsingError("The only possible keyword before a 'def' is 'async', not '%s'" % async_maybe["value"])

        return [{
            "type": "def",
            "async": async_maybe["async"],
            "return_annotation": return_annotation["value"],
            "return_annotation_first_formatting": return_annotation["first_formatting"],
            "return_annotation_second_formatting": return_annotation["second_formatting"],
            "async_formatting": async_maybe.get("formatting", []),
            "decorators": [],
            "name": name.value,
            "first_formatting": def_.hidden_tokens_after,
            "second_formatting": left_parenthesis.hidden_tokens_before,
            "third_formatting": left_parenthesis.hidden_tokens_after,
            "fourth_formatting": right_parenthesis.hidden_tokens_before,
            "fifth_formatting": colon.hidden_tokens_before,
            "sixth_formatting": colon.hidden_tokens_after,
            "arguments": typed_parameters,
            "value": suite,
        }]

    @pg.production("return_annotation : ")
    def return_annotation_empty(pack):
        return {
            "value": {},
            "first_formatting": [],
            "second_formatting": [],
        }

    @pg.production("return_annotation : RIGHT_ARROW test")
    def return_annotation(pack):
        right_arrow, test = pack
        return {
            "value": test,
            "first_formatting": right_arrow.hidden_tokens_before,
            "second_formatting": right_arrow.hidden_tokens_after,
        }

    @pg.production("argslist : argslist argument")
    @pg.production("testlist_argslist : argslist argument")
    @pg.production("typed_parameters : typed_parameters typed_parameter")
    @pg.production("parameters : parameters parameter")
    def parameters_parameters_parameter(pack,):
        (parameters, parameter,) = pack
        return parameters + parameter

    @pg.production("argslist : argument")
    @pg.production("testlist_argslist : argument")
    @pg.production("typed_parameters : typed_parameter")
    @pg.production("parameters : parameter")
    def parameters_parameter(pack,):
        (parameter,) = pack
        return parameter

    @pg.production("argument :")
    @pg.production("typed_parameter : ")
    @pg.production("parameter : ")
    def parameter_empty(p):
        return []

    @pg.production("name : NAME")
    def name(pack):
        (name_,) = pack
        return {
            "type": "name",
            "value": name_.value,
        }

    @pg.production("typed_parameter : LEFT_PARENTHESIS name RIGHT_PARENTHESIS maybe_test")
    @pg.production("parameter : LEFT_PARENTHESIS name RIGHT_PARENTHESIS maybe_test")
    def parameter_fpdef(pack):
        (left_parenthesis, name, right_parenthesis, (equal, test)) = pack
        return [{
            "type": "def_argument",
            "annotation": {},
            "annotation_first_formatting": [],
            "annotation_second_formatting": [],
            "first_formatting": equal.hidden_tokens_before if equal else [],
            "second_formatting": equal.hidden_tokens_after if equal else [],
            "value": test,
            "target": {
                "type": "associative_parenthesis",
                "first_formatting": left_parenthesis.hidden_tokens_before,
                "second_formatting": left_parenthesis.hidden_tokens_after,
                "third_formatting": right_parenthesis.hidden_tokens_before,
                "fourth_formatting": right_parenthesis.hidden_tokens_after,
                "value": name
            }
        }]

    @pg.production("typed_parameter : LEFT_PARENTHESIS fplist RIGHT_PARENTHESIS maybe_test")
    @pg.production("parameter : LEFT_PARENTHESIS fplist RIGHT_PARENTHESIS maybe_test")
    def parameter_fplist(pack):
        (left_parenthesis, fplist, right_parenthesis, (equal, test)) = pack
        return [{
            "type": "def_argument",
            "annotation": {},
            "annotation_first_formatting": [],
            "annotation_second_formatting": [],
            "first_formatting": equal.hidden_tokens_before if equal else [],
            "second_formatting": equal.hidden_tokens_after if equal else [],
            "value": test,
            "target": {
                "type": "tuple",
                "with_parenthesis": True,
                "first_formatting": left_parenthesis.hidden_tokens_after,
                "second_formatting": left_parenthesis.hidden_tokens_before,
                "third_formatting": right_parenthesis.hidden_tokens_before,
                "fourth_formatting": right_parenthesis.hidden_tokens_after,
                "value": fplist,
            },
        }]

    @pg.production("fplist : fplist parameter")
    def fplist_recur(pack):
        (fplist, name) = pack
        if name[0]["type"] == "def_argument":
            name = [name[0]["target"]]
        return fplist + name

    @pg.production("fplist : parameter comma")
    def fplist(pack):
        (name, comma) = pack
        if name[0]["type"] == "def_argument":
            name = [name[0]["target"]]
        return name + [comma]

    # really strange that left part of argument grammar can be a test
    # I guess it's yet another legacy mistake
    # python give me 'SyntaxError: keyword can't be an expression' when I try to
    # put something else than a name (looks like a custom SyntaxError)
    @pg.production("argument : test maybe_test")
    def named_argument(pack):
        (name, (equal, test)) = pack
        return [{
            "type": "call_argument",
            "first_formatting": equal.hidden_tokens_before if equal else [],
            "second_formatting": equal.hidden_tokens_after if equal else [],
            "value": test if equal else name,
            "target": name if equal else {}
        }]

    @pg.production("typed_parameter : name COLON test maybe_test")
    def parameter_annotation_with_default(pack):
        # name, (equal, test) = pack
        name, colon, annotation, (equal, test) = pack
        return [{
            "type": "def_argument",
            "annotation": annotation,
            "annotation_first_formatting": colon.hidden_tokens_before,
            "annotation_second_formatting": colon.hidden_tokens_after,
            "first_formatting": equal.hidden_tokens_before if equal else [],
            "second_formatting": equal.hidden_tokens_after if equal else [],
            "value": test,
            "target": name
        }]

    @pg.production("typed_parameter : name maybe_test")
    def parameter_alone_with_default(pack):
        name, (equal, test) = pack
        return [{
            "type": "def_argument",
            "annotation": {},
            "annotation_first_formatting": [],
            "annotation_second_formatting": [],
            "first_formatting": equal.hidden_tokens_before if equal else [],
            "second_formatting": equal.hidden_tokens_after if equal else [],
            "value": test,
            "target": name
        }]

    @pg.production("parameter : name maybe_test")
    def parameter_with_default(pack):
        name, (equal, test) = pack
        return [{
            "type": "def_argument",
            "annotation": {},
            "annotation_first_formatting": [],
            "annotation_second_formatting": [],
            "first_formatting": equal.hidden_tokens_before if equal else [],
            "second_formatting": equal.hidden_tokens_after if equal else [],
            "value": test,
            "target": name
        }]

    @pg.production("maybe_test : EQUAL test")
    def maybe_test(pack):
        return pack

    @pg.production("maybe_test : ")
    def maybe_test_empty(pack):
        return (None, {})

    @pg.production("argument : test comp_for")
    def generator_comprehension(pack):
        (test, comp_for,) = pack
        return [{
            "type": "argument_generator_comprehension",
            "result": test,
            "generators": comp_for,
        }]

    @pg.production("argument : STAR test")
    def argument_star(pack):
        (star, test,) = pack
        return [{
            "type": "list_argument",
            "annotation": {},
            "annotation_first_formatting": [],
            "annotation_second_formatting": [],
            "formatting": star.hidden_tokens_after,
            "value": test,
        }]

    @pg.production("argument : DOUBLE_STAR test")
    def argument_star_star(pack):
        (double_star, test,) = pack
        return [{
            "type": "dict_argument",
            "annotation": {},
            "annotation_first_formatting": [],
            "annotation_second_formatting": [],
            "formatting": double_star.hidden_tokens_after,
            "value": test,
        }]

    @pg.production("typed_parameter : STAR NAME COLON test")
    def typed_parameter_star(pack):
        (star, name, colon, test) = pack
        return [{
            "type": "list_argument",
            "formatting": star.hidden_tokens_after,
            "annotation": test,
            "annotation_first_formatting": colon.hidden_tokens_before if colon else [],
            "annotation_second_formatting": colon.hidden_tokens_after if colon else [],
            "value": {
                "type": "name",
                "value": name.value,
            }
        }]

    @pg.production("typed_parameter : DOUBLE_STAR NAME COLON test")
    def typed_parameter_double_star(pack):
        (double_star, name, colon, test) = pack
        return [{
            "type": "dict_argument",
            "formatting": double_star.hidden_tokens_after,
            "annotation": test,
            "annotation_first_formatting": colon.hidden_tokens_before if colon else [],
            "annotation_second_formatting": colon.hidden_tokens_after if colon else [],
            "value": {
                "type": "name",
                "value": name.value,
            }
        }]

    # TODO refactor those 2 to standardize with argument_star and argument_star_star
    @pg.production("typed_parameter : STAR NAME")
    @pg.production("parameter : STAR NAME")
    def parameter_star(pack):
        (star, name,) = pack
        return [{
            "type": "list_argument",
            "annotation": {},
            "annotation_first_formatting": [],
            "annotation_second_formatting": [],
            "formatting": star.hidden_tokens_after,
            "value": {
                "type": "name",
                "value": name.value,
            }
        }]

    # TODO refactor those 2 to standardize with argument_star and argument_star_star
    @pg.production("typed_parameter : STAR")
    @pg.production("parameter : STAR")
    def parameter_star_only(pack):
        (star, ) = pack
        return [{
            "type": "kwargs_only_marker",
            "formatting": star.hidden_tokens_after,
        }]

    @pg.production("typed_parameter : DOUBLE_STAR NAME")
    @pg.production("parameter : DOUBLE_STAR NAME")
    def parameter_star_star(pack):
        (double_star, name,) = pack
        return [{
            "type": "dict_argument",
            "annotation": {},
            "annotation_first_formatting": [],
            "annotation_second_formatting": [],
            "formatting": double_star.hidden_tokens_after,
            "value": {
                "type": "name",
                "value": name.value,
            },
        }]

    @pg.production("argument : comma")
    @pg.production("typed_parameter : comma")
    @pg.production("parameter : comma")
    def parameter_comma(pack):
        (comma,) = pack
        return [comma]

    @pg.production("suite : simple_stmt")
    def suite(pack):
        (simple_stmt,) = pack
        return simple_stmt

    @pg.production("suite : endls INDENT statements DEDENT")
    def suite_indent(pack):
        (endls, indent, statements, dedent,) = pack
        return endls + statements

    @pg.production("endls : endls endl")
    @pg.production("endls : endl")
    def endls(p):
        if len(p) == 1:
            return p[0]
        return p[0] + p[1]

    include_imports(pg)
    include_control_structures(pg)
    include_primivites(pg, print_function)
    include_operators(pg)
    include_data_structures(pg)

    @pg.production("atom : LEFT_PARENTHESIS yield_expr RIGHT_PARENTHESIS")
    def yield_atom(pack):
        (left_parenthesis, yield_expr, right_parenthesis) = pack
        return {
            "type": "yield_atom",
            "value": yield_expr["value"],
            "first_formatting": left_parenthesis.hidden_tokens_after,
            "second_formatting": yield_expr["formatting"],
            "third_formatting": right_parenthesis.hidden_tokens_before
        }

    @pg.production("atom : BACKQUOTE testlist1 BACKQUOTE")
    def repr_atom(pack):
        (backquote, testlist1, backquote2) = pack
        return {
            "type": "repr",
            "value": testlist1,
            "first_formatting": backquote.hidden_tokens_after,
            "second_formatting": backquote2.hidden_tokens_before,
        }

    @pg.production("testlist1 : test comma testlist1")
    def testlist1_double(pack):
        (test, comma, test2,) = pack
        return [test, comma] + test2

    @pg.production("testlist1 : test")
    def testlist1(pack):
        (test,) = pack
        return [test]

    # TODO test all the things (except INT)
    @pg.production("atom : INT")
    @pg.production("atom : LONG")
    @pg.production("atom : OCTA")
    @pg.production("atom : HEXA")
    @pg.production("atom : BINARY")
    @pg.production("atom : FLOAT")
    @pg.production("atom : FLOAT_EXPONANT")
    @pg.production("atom : FLOAT_EXPONANT_COMPLEX")
    @pg.production("atom : COMPLEX")
    def int(pack):
        (int_,) = pack
        return create_node_from_token(int_, section="number")

    @pg.production("atom : name")
    def atom_name(pack):
        (name,) = pack
        return name

    @pg.production("atom : strings")
    def strings(pack):
        (string_chain,) = pack
        if len(string_chain) == 1:
            return string_chain[0]
        return {
            "type": "string_chain",
            "value": string_chain
        }

    @pg.production("strings : string strings")
    def strings_string_strings(pack):
        (string_, strings_) = pack
        return string_ + strings_

    @pg.production("strings : string")
    def strings_string(pack):
        (string_,) = pack
        return string_

    # TODO tests those other kind of strings
    @pg.production("string : STRING")
    @pg.production("string : RAW_STRING")
    @pg.production("string : INTERPOLATED_STRING")
    @pg.production("string : UNICODE_STRING")
    @pg.production("string : BINARY_STRING")
    @pg.production("string : UNICODE_RAW_STRING")
    @pg.production("string : BINARY_RAW_STRING")
    @pg.production("string : INTERPOLATED_RAW_STRING")
    def string(pack):
        (string_,) = pack
        return [{
            "type": string_.name.lower(),
            "value": string_.value,
            "first_formatting": string_.hidden_tokens_before,
            "second_formatting": string_.hidden_tokens_after,
        }]

    @pg.production("comma : COMMA")
    def comma(pack):
        (comma,) = pack
        return {
            "type": "comma",
            "first_formatting": comma.hidden_tokens_before,
            "second_formatting": comma.hidden_tokens_after,
        }

    def parse(tokens):
        if print_function:
            new_tokens = []
            for token in tokens:
                if token[0] in ("PRINT", "EXEC"):
                    token = list(token)
                    token[0] = "NAME"
                    token = tuple(token)
                new_tokens.append(token)

            tokens = [BaronToken(*x) if x else x for x in new_tokens] + [None]
        else:
            tokens = [BaronToken(*x) if x else x for x in tokens] + [None]

        return parser.parse(iter(tokens))

    parser = pg.build()
    return parse


def fake_lexer(sequence):
    for i in tokenize(sequence):
        if i is None:
            yield None
        yield BaronToken(*i)


def parse(sequence):
    parser = generate_parse(print_function=False)
    return parser.parse(fake_lexer(sequence))



from .render import RenderWalker, child_by_key
from .utils import is_newline, split_on_newlines, total_ordering
from copy import deepcopy


def position_to_path(tree, position):
    """Path to the node located at the given line and column

    This function locates a node in the rendered source code
    """
    return PositionFinder().find(tree, position)


def path_to_node(tree, path):
    """FST node located at the given path"""
    if path is None:
        return None

    node = tree

    for key in path:
        node = child_by_key(node, key)

    return node


def position_to_node(tree, position):
    """FST node located at the given line and column"""
    return path_to_node(tree, position_to_path(tree, position))


def node_to_bounding_box(node):
    """Bounding box of the given node

    The bounding box of a node represents its left most and right most
    position in the rendered source code. Its left position is here
    always (1, 1).
    """
    return BoundingBoxFinder().compute(node)


def path_to_bounding_box(tree, path):
    """Absolute bounding box of the node located at the given path"""
    return BoundingBoxFinder().compute(tree, path)


@total_ordering
class Position(object):
    """Handles a cursor's line and column

    Operations requiring another Position as argument can be given an
    indexable object of len >= 2 where the index 0 contains the line and
    the index 1 contains the column. For example a tuple of len 2.
    """
    def __init__(self, position):
        if hasattr(position, 'line') and hasattr(position, 'column'):
            self.line = position.line
            self.column = position.column
        elif len(position) >= 2:
            self.line = position[0]
            self.column = position[1]
        else:
            raise AttributeError(position)

    def advance_columns(self, columns):
        """(3, 10) -> (3, 11)"""
        self.column += columns

    def advance_line(self):
        """(3, 10) -> (4, 1)"""
        self.line += 1
        self.column = 1

    @property
    def left(self):
        """(3, 10) -> (3, 9)"""
        return Position((self.line, self.column - 1))

    @property
    def right(self):
        """(3, 10) -> (3, 11)"""
        return Position((self.line, self.column + 1))

    def __add__(self, other):
        """(1, 1) + (1, 1) -> (2, 2)"""
        other = Position(other)
        return Position((self.line + other.line,
                        self.column + other.column))

    def __neg__(self):
        """(1, -1) -> (-1, 1)"""
        return Position((-self.line, -self.column))

    def __sub__(self, other):
        """(1, 1) - (1, 1) -> (0, 0)"""
        other = Position(other)
        return Position((self.line - other.line,
                        self.column - other.column))

    def __nonzero__(self):
        return self.line >= 0 and self.column >= 0

    def __bool__(self):
        return self.__nonzero__()

    def __eq__(self, other):
        """Compares Positions or Position and tuple

        Will not fail if other is an unsupported type"""
        if not (hasattr(other, 'line') and hasattr(other, 'column')) and len(other) < 2:
            return False

        other = Position(other)
        return self.line == other.line and self.column == other.column

    def __lt__(self, other):
        """Compares Position with Position or indexable object"""
        other = Position(other)
        return (self.line, self.column) < (other.line, other.column)

    def __repr__(self):
        return 'Position (%s, %s)' % (str(self.line), str(self.column))

    def to_tuple(self):
        return (self.line, self.column)


class BoundingBox(object):
    """Handles a selection's top_left and bottom_right position

    Operations requiring another BoundingBox as argument can be given an
    indexable object of len >= 2 where the index 0 contains the top_left
    position, either as a Position or an indexable object. The index
    1 must contain, in a similar manner, the bottom_right position.
    """
    def __init__(self, bounding_box):
        if hasattr(bounding_box, 'top_left') and hasattr(bounding_box, 'bottom_right'):
            self.top_left = Position(bounding_box.top_left)
            self.bottom_right = Position(bounding_box.bottom_right)
        elif len(bounding_box) >= 2:
            self.top_left = Position(bounding_box[0])
            self.bottom_right = Position(bounding_box[1])
        else:
            raise AttributeError(bounding_box)

    def __eq__(self, other):
        """Compares BoundingBox with BoundingBox or indexable object"""
        if not (hasattr(other, 'top_left') and hasattr(other, 'bottom_right')) and len(other) < 2:
            return False

        other = BoundingBox(other)
        return self.top_left == other.top_left and self.bottom_right == other.bottom_right

    def __repr__(self):
        return 'BoundingBox (%s, %s)' % (str(self.top_left), str(self.bottom_right))


class PathWalker(RenderWalker):
    """Gives the current path while walking the rendered tree

    It adds an attribute "current_path" which is updated each time the
    walker takes a step.
    """
    def walk(self, tree):
        self.current_path = []

        super(PathWalker, self).walk(tree)

    def before(self, key_type, item, render_key):
        if render_key is not None:
            self.current_path.append(render_key)

        return super(PathWalker, self).before(key_type, item, render_key)

    def after(self, key_type, item, render_key):
        stop = super(PathWalker, self).after(key_type, item, render_key)

        if render_key is not None:
            self.current_path.pop()

        return stop


class PositionFinder(PathWalker):
    """Find a node by line and column and return the path to it.

    First, walk through all the nodes while maintaining the current line
    and column. When the targetted node is found, stop there and build
    the path while going back up the tree.
    """
    def find(self, tree, position):
        self.current = Position((1, 1))
        self.target = Position(position)
        self.found_path = None

        self.walk(tree)
        return self.found_path

    def before_constant(self, constant, key):
        """Determine if we're on the targetted node.

        If the targetted column is reached, `stop` and `path_found` are
        set. If the targetted line is passed, only `stop` is set. This
        prevents unnecessary tree travelling when the targetted column
        is out of bounds.
        """
        newlines_split = split_on_newlines(constant)

        for c in newlines_split:
            if is_newline(c):
                self.current.advance_line()
                # if target line is passed
                if self.current.line > self.target.line:
                    return self.STOP

            else:
                advance_by = len(c)
                if self.is_on_targetted_node(advance_by):
                    self.found_path = deepcopy(self.current_path)
                    return self.STOP
                self.current.advance_columns(advance_by)

    before_string = before_constant

    def is_on_targetted_node(self, advance_by):
        return self.target.line == self.current.line \
            and self.target.column >= self.current.column \
            and self.target.column < self.current.column + advance_by


class BoundingBoxFinder(PathWalker):
    """Compute the bounding box of the given node.

    First, walk to the target path while incrementing the position.
    When reached, the top-left position is set to the current position.
    Then walk the whole node, still incrementing the position. When
    arriving at the end of the node, store the previous position, not
    the current one, as the bottom-right position.
    If no target path is given, assume the targetted node is the whole
    tree.
    """
    def compute(self, tree, target_path=None):
        self.target_path = target_path
        self.current_position = Position((1, 1))
        self.left_of_current_position = Position((1, 0))
        self.top_left = None
        self.bottom_right = None
        self.found = True if self.target_path is None or len(target_path) == 0 else False

        self.walk(tree)

        if self.found and self.top_left is None and self.bottom_right is None:
            return BoundingBox((Position((1, 1)), self.left_of_current_position))

        return BoundingBox((self.top_left, self.bottom_right))

    def before(self, key_type, item, render_key):
        stop = super(BoundingBoxFinder, self).before(key_type, item, render_key)

        if self.current_path == self.target_path:
            self.found = True
            self.top_left = deepcopy(self.current_position)

        if key_type not in ['constant', 'string']:
            return stop

        newlines_split = split_on_newlines(item)

        for c in newlines_split:
            if is_newline(c):
                self.current_position.advance_line()
                self.left_of_current_position = self.current_position.left
            elif c != "":
                self.current_position.advance_columns(len(c))
                self.left_of_current_position = self.current_position.left

        return stop

    def after(self, key_type, item, render_key):
        if self.bottom_right is None and self.found and self.current_path == self.target_path:
            self.bottom_right = deepcopy(self.left_of_current_position)

        return super(BoundingBoxFinder, self).after(key_type, item, render_key)



from .spliter import split
from .grouper import group
from .tokenizer import tokenize as _tokenize
from .formatting_grouper import group as space_group
from .future import has_print_function, replace_print_by_name
from .grammator import generate_parse
from .indentation_marker import mark_indentation
from .inner_formatting_grouper import group as inner_group
from .parser import ParsingError


parse_tokens = generate_parse(False)
parse_tokens_print_function = generate_parse(True)


def _parse(tokens, print_function):
    parser = parse_tokens if not print_function else parse_tokens_print_function
    try:
        try:
            return parser(tokens)
        except ParsingError:
            # swap parsers for print_function situation where I failed to find it
            parser = parse_tokens if print_function else parse_tokens_print_function
            return parser(tokens)
    except ParsingError as e:
        raise
    except Exception as e:
        import sys
        import traceback
        traceback.print_exc(file=sys.stderr)
        sys.stderr.write("%s\n" % e)
        sys.stderr.write("\nBaron has failed to parse this input. If this is valid python code (and by that I mean that the python binary successfully parse this code without any syntax error) (also consider that python does not yet parse python 3 code integrally) it would be kind if you can extract a snippet of your code that make Baron fails and open a bug here: https://github.com/PyCQA/baron/issues\n\nSorry for the inconvenience.")


def parse(source_code, print_function=None):
    # Python syntax requires source code to end with an ENDL token
    # the endl token is removed afterward if and only if it's the last token of the root level
    # It is possible that this token end up in a 'suite' grammar rule
    # which means that it is 'traped' in an indented block of code
    # I don't want to recursively cross the tree to hope to find it
    # This solution behave in the expected way for 90% of the case
    newline_appended = False
    linesep = "\r\n" if source_code.endswith("\r\n") else "\n"
    if source_code and not source_code.endswith(linesep):
        source_code += linesep
        newline_appended = True

    if print_function is None:
        tokens = tokenize(source_code, False)
        print_function = has_print_function(tokens)
        if print_function:
            replace_print_by_name(tokens)
    else:
        tokens = tokenize(source_code, print_function)

    if newline_appended:
        to_return = _parse(tokens, print_function)

        if to_return[-1]["type"] == "endl" and not to_return[-1]["formatting"]:
            return to_return[:-1]
        elif to_return[-1]["type"] == "endl" and to_return[-1]["formatting"]:
            return to_return[:-1] + to_return[-1]["formatting"]
        else:
            return to_return

    return _parse(tokens, print_function)


def tokenize(pouet, print_function=False):
    splitted = split(pouet)
    grouped = group(splitted)
    print_tokenized = _tokenize(grouped, print_function)
    space_grouped = space_group(print_tokenized)
    inner_grouped = inner_group(space_grouped)
    indentation_marked = mark_indentation(inner_grouped)
    return indentation_marked
